""" Terminal output package for x/84. """
import warnings
import inspect
import random
import glob
import os
import re

# local
from x84.bbs.ini import get_ini
from x84.bbs.session import getterminal, getsession

# 3rd-party
from sauce import SAUCE

#: A mapping of SyncTerm fonts/code pages to their sequence value, for use
#: as argument ``font_name`` of :func:`syncterm_setfont`.
#:
#: Where matching, their python-standard encoding value is used, (fe. 'cp437').
#: Otherwise, the lower-case named of the font is used.
#:
#: source:
#: http://cvs.synchro.net/cgi-bin/viewcvs.cgi/*checkout*/src/conio/cterm.txt
SYNCTERM_FONTMAP = (
    "cp437", "cp1251", "koi8_r", "iso8859_2", "iso8859_4", "cp866",
    "iso8859_9", "haik8", "iso8859_8", "koi8_u", "iso8859_15", "iso8859_4",
    "koi8_r_b", "iso8859_4", "iso8859_5", "ARMSCII_8", "iso8859_15",
    "cp850", "cp850", "cp885", "cp1251", "iso8859_7", "koi8-r_c",
    "iso8859_4", "iso8859_1", "cp866", "cp437", "cp866", "cp885",
    "cp866_u", "iso8859_1", "cp1131", "c64_upper", "c64_lower",
    "c128_upper", "c128_lower", "atari", "pot_noodle", "mo_soul",
    "microknight", "topaz",)

#: Translation map for embedded font hints in SAUCE records as documented at
#: http://www.acid.org/info/sauce/sauce.htm section FontName. Used by
#: :func:`showart` to automatically determine which codepage to be used
#: by utf8 terminals to provide an approximate translation.
SAUCE_FONT_MAP = {
    'Amiga MicroKnight': 'amiga',
    'Amiga MicroKnight+': 'amiga',
    'Amiga mOsOul': 'amiga',
    'Amiga P0T-NOoDLE': 'amiga',
    'Amiga Topaz 1': 'amiga',
    'Amiga Topaz 1+': 'amiga',
    'Amiga Topaz 2': 'amiga',
    'Amiga Topaz 2+': 'amiga',
    'Atari ATASCII': 'atari',
    'IBM EGA43': 'cp437',
    'IBM EGA': 'cp437',
    'IBM VGA25G': 'cp437',
    'IBM VGA50': 'cp437',
    'IBM VGA': 'cp437',
}

# IBM-PC code pages
for page in (
    '437', '720', '737', '775', '819', '850', '852', '855', '857', '858',
    '860', '861', '862', '863', '864', '865', '866', '869', '872',
):
    codec = 'cp%s' % (page,)
    SAUCE_FONT_MAP.update({
        'IBM EGA43 %s' % (page,): codec,
        'IBM EGA %s' % (page,): codec,
        'IBM VGA25g %s' % (page,): codec,
        'IBM VGA50 %s' % (page,): codec,
        'IBM VGA %s' % (page,): codec,
    })

#: simple regular expression for matching simple ansi colors,
#: for use by :func:`encode_pipe`.
RE_ANSI_COLOR = re.compile(r'\033\[(\d{2,3})m')


def syncterm_setfont(font_name, font_page=0):
    """
    Send SyncTerm's sequence for selecting a "font" codepage.

    :param str font_name: any value of :py:const:`SYNCTERM_FONTMAP`.
    :param int font_page:

    Reference::

        CSI [ p1 [ ; p2 ] ] sp D
        Font Selection
        Defaults: p1 = 0  p2 = 0
        "sp" indicates a single space character.
        Sets font p1 to be the one indicated by p2.  Currently only the primary
        font (Font zero) and secondary font (Font one) are supported.  p2 must
        be between 0 and 255.  Not all output types support font selection.
        Only X11 and SDL currently do.

    source:
    http://cvs.synchro.net/cgi-bin/viewcvs.cgi/*checkout*/src/conio/cterm.txt

    """
    # font_code is the index of font_name in SYNCTERM_FONTMAP, so that
    # syncterm_setfont('topaz') becomes value 40, and returns sequence
    # "\x1b[0;40 D'
    try:
        font_code = SYNCTERM_FONTMAP.index(font_name)
    except IndexError:
        raise ValueError("The specified font_name={0!r} is not any of the "
                         "available fonts specified in module {1}, table "
                         "SYNCTERM_FONTMAP. Available values: {2!r}".format(
                             font_name, __name__, SYNCTERM_FONTMAP))
    return u'\x1b[{0};{1} D'.format(font_page, font_code)


def echo(ucs):
    """
    Display unicode terminal sequence.

    :param str ucs: unicode sequence to write to terminal.
    """
    session = getsession()
    if not isinstance(ucs, unicode):
        warnings.warn('non-unicode: %r' % (ucs,), UnicodeWarning, 2)
        return session.write(ucs.decode('iso8859-1'))
    return session.write(ucs)


def timeago(secs, precision=0):
    """
    Return human-readable string of seconds elapsed.

    :param int secs: number of seconds "ago".
    :param int precision: optional decimal precision of returned seconds.

    Pass a duration of time and return human readable shorthand, fe::

        >>> asctime(126.32)
        ' 2m 6s',
        >>> asctime(70.9999, 2)
        ' 1m 10.99s'
    """
    # split by days, mins, hours, secs
    years = weeks = days = mins = hours = 0
    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    days, hours = divmod(hours, 24)
    weeks, days = divmod(days, 7)
    years, weeks = divmod(weeks, 52)
    ((num1, num2), (label1, label2)) = (
        ((years, weeks), (u'y', u'w')) if years >= 1.0 else
        ((weeks, days), (u'w', u'd')) if weeks >= 1.0 else
        ((days, hours), (u'd', u'h')) if days >= 1.0 else
        ((hours, mins), (u'h', u'm')) if hours >= 1.0 else
        ((mins, secs), (u'm', u's')))
    return (u'%2d%s%2.*f%s' % (num1, label1, precision, num2, label2,))


def decode_pipe(ucs):
    """
    Return ucs containing 'pipe codes' with terminal color sequences.

    These are sometimes known as LORD codes, as they were used in the DOS Door
    game of the same name. Compliments :func:`encode_pipe`.

    :param str ucs: string containing 'pipe codes'.
    :rtype: str
    """
    # simple optimization, no '|' ? exit early!
    if u'|' not in ucs:
        return ucs
    term = getterminal()
    outp = u''
    ptr = 0
    match = None
    ANSI_PIPE = re.compile(r'\|(\d{2,3}|\|)')

    for match in ANSI_PIPE.finditer(ucs):
        val = match.group(1)
        # allow escaping using a second pipe
        if val == u'|':
            outp += ucs[ptr:match.start() + 1]
            ptr = match.end()
            continue
        # 0..7 -> 7
        while val.startswith('0'):
            val = val[1:]
        int_value = 0 if 0 == len(val) else int(val, 10)
        assert int_value >= 0 and int_value <= 256
        # colors 0-7 and 16-256 are as-is term.color()
        # special accommodations for 8-15, some termcaps are ok
        # with term.color(11), whereas others have trouble, help
        # out by using dim color and bold attribute instead.
        attr = u''
        if int_value == 7:
            attr = term.normal
        elif int_value < 7 or int_value >= 16:
            attr = term.normal + term.color(int_value)
        elif int_value <= 15:
            attr = term.normal + term.color(int_value - 8) + term.bold
        outp += ucs[ptr:match.start()] + attr
        ptr = match.end()

    outp = ucs if match is None else u''.join((outp, ucs[match.end():]))
    return u''.join((outp, term.normal))


def encode_pipe(ucs):
    """
    Given a string containing ECMA-48 sequence, replace with "pipe codes".

    These are sometimes known as LORD codes, as they were used in the DOS Door
    game of the same name. Compliments :func:`decode_pipe`.

    :param str ucs: string containing ECMA-48 sequences.
    :rtype: str
    """
    # (dead url ...)
    # http://wiki.mysticbbs.com/mci_codes
    #
    # TODO: Support all kinds of terminal color sequences,
    # such as kermit or avatar or some such, something non-emca,
    # upstream blessed project is looking for a SequenceIterator
    # class, https://github.com/jquast/blessed/issues/29
    outp = u''
    nxt = 0
    for idx in range(0, len(ucs)):
        if idx == nxt:
            # at sequence, point beyond it,
            match = RE_ANSI_COLOR.match(ucs[idx:])
            if match:
                nxt = idx + len(match.group(0))
                value = int(match.group(1)) - 30
                if value >= 0 and value <= 60:
                    outp += u'|%02d' % (value,)
        if nxt <= idx:
            # append non-sequence to outp,
            outp += ucs[idx]
            # point beyond next sequence, if any,
            # otherwise point to next character
            nxt = idx + 1
    return outp


def ropen(filename, mode='rb'):
    """ Open random file using wildcard (glob). """
    files = glob.glob(filename)
    return open(random.choice(files), mode) if len(files) else None


def showart(filepattern, encoding=None, auto_mode=True, center=False,
            poll_cancel=False, msg_cancel=None, force=False):
    """
    Yield unicode sequences for any given ANSI Art (of art_encoding).

    Effort is made to parse SAUCE data, translate input to unicode, and trim
    artwork too large to display.  If ``poll_cancel`` is not ``False``,
    represents time as float for each line to block for keypress -- if any is
    received, then iteration ends and ``msg_cancel`` is displayed as last line
    of art.

    If you provide no ``encoding``, the piece encoding will be based on either
    the encoding in the SAUCE record, the configured default or the default
    fallback ``CP437`` encoding.

    Alternate codecs are available if you provide the ``encoding`` argument.
    For example, if you want to show an Amiga style ASCII art file::

        >>> from x84.bbs import echo, showart
        >>> for line in showart('test.asc', 'topaz'):
        ...     echo(line)

    The ``auto_mode`` flag will, if set, only respect the selected encoding if
    the active session is UTF-8 capable.

    If ``center`` is set to ``True``, the piece will be centered respecting the
    current terminal's width.

    If ``force`` is set to true then the artwork will be displayed even if it's
    wider than the screen.

    """
    # pylint: disable=R0913,R0914
    #         Too many arguments
    #         Too many local variables
    term = getterminal()

    # When the given artfile pattern's folder is not absolute, nor relative to
    # our cwd, build a relative position of the folder by the calling module's
    # containing folder.  This only works for subdirectories (like 'art/').
    _folder = os.path.dirname(filepattern)
    if not (_folder.startswith(os.path.sep) or os.path.isdir(_folder)):
        # On occasion, after a general exception in a script, re-calling the
        # same script may cause yet another exception, HERE.  The 2nd call is
        # fine though; this only would effect a developer.
        #
        # Just try again.
        caller_module = inspect.stack()[1][1]
        rel_folder = os.path.dirname(caller_module)
        if _folder:
            rel_folder = os.path.join(rel_folder, _folder)
        if os.path.isdir(rel_folder):
            filepattern = os.path.join(
                rel_folder,
                os.path.basename(filepattern))

    # Open the piece
    try:
        filename = os.path.relpath(random.choice(glob.glob(filepattern)))
    except IndexError:
        filename = None

    if filename is None:
        yield u''.join((
            term.bold_red(u'-- '),
            u'no files matching {0}'.format(filepattern),
            term.bold_red(u' --'),
        ))
        return

    file_basename = os.path.basename(filename)

    # Parse the piece
    parsed = SAUCE(filename)

    # If no explicit encoding is given, we go through a couple of steps to
    # resolve the possible file encoding:
    if encoding is None:
        # 1. See if the SAUCE record has a font we know about, it's in the
        #    filler
        if parsed.record and parsed.filler_str in SAUCE_FONT_MAP:
            encoding = SAUCE_FONT_MAP[parsed.filler_str]

        # 2. Get the system default art encoding,
        #    or fall-back to cp437
        else:
            encoding = get_ini('system', 'art_utf8_codec') or 'cp437'

    # If auto_mode is enabled, we'll only use the input encoding on UTF-8
    # capable terminals, because our codecs do not know how to "transcode"
    # between the various encodings.
    if auto_mode:
        def _decode(what):
            # pylint: disable=C0111
            #         Missing function docstring (col 8)
            session = getsession()
            if session.encoding == 'utf8':
                return what.decode(encoding)
            elif session.encoding == 'cp437':
                return what.decode('cp437')
            else:
                return what

    # If auto_mode is disabled, we'll just respect whatever input encoding was
    # selected before
    else:
        _decode = lambda what: what.decode(encoding)

    # For wide terminals, center piece on screen using cursor movement
    # when center=True.
    padding = u''
    if center and term.width > 81:
        padding = term.move_x((term.width / 2) - 40)
    lines = _decode(parsed.data).splitlines()
    for idx, line in enumerate(lines):

        if poll_cancel is not False and term.inkey(poll_cancel):
            # Allow slow terminals to cancel by pressing a keystroke
            msg_cancel = msg_cancel or u''.join(
                (term.normal,
                 term.bold_black(u'-- '),
                 u'canceled {0} by input'.format(os.path.basename(filename)),
                 term.bold_black(u' --'),
                 ))
            yield u'\r\n' + term.center(msg_cancel).rstrip() + u'\r\n'
            return

        line_length = term.length(line.rstrip())

        if force is False and not padding and term.width < line_length:
            # if the artwork is too wide and force=False, simply stop displaying it.
            msg_too_wide = u''.join(
                (term.normal,
                 term.bold_black(u'-- '),
                 (u'canceled {0}, too wide:: {1}'
                  .format(file_basename, line_length)),
                 term.bold_black(u' --'),
                 ))
            yield (u'\r\n' +
                   term.center(msg_too_wide).rstrip() +
                   u'\r\n')
            return
        if idx == len(lines) - 1:
            # strip DOS end of file (^Z)
            line = line.rstrip('\x1a')
            if not line.strip():
                break
        yield padding + line + u'\r\n'
    yield term.normal


def from_cp437(text):
    """ Deprecated form of ``bytes.decode('cp437_art')``. """
    warnings.warn('from_cp437() is deprecated, use bytes.decode("cp437_art")')
    return text.decode('cp437_art')



""" Utility functions for clients based on telnetlib for x/84. """

SEND = chr(1)  # Sub-process negotiation SEND command


def callback_cmdopt(socket, cmd, opt, env_term=None, width=None, height=None):
    """ Callback for telnetlib.Telnet.set_option_negotiation_callback. """
    # pylint: disable=R0913
    #         Too many arguments
    import telnetlib
    import struct
    IS = chr(0)  # Sub-process negotiation IS command
    if env_term is None:
        env_term = 'vt220'
    width = width or 80
    height = height or 24
    if cmd == telnetlib.WILL:
        if opt in (telnetlib.ECHO, telnetlib.SGA):
            socket.sendall(telnetlib.IAC + telnetlib.DO + opt)
    elif cmd == telnetlib.DO:
        if opt == telnetlib.SGA:
            socket.sendall(telnetlib.IAC + telnetlib.WILL + opt)
        elif opt == telnetlib.TTYPE:
            socket.sendall(telnetlib.IAC + telnetlib.WILL + opt)
            socket.sendall(telnetlib.IAC + telnetlib.SB
                           + telnetlib.TTYPE + IS + env_term
                           + chr(0) + telnetlib.IAC + telnetlib.SE)
        elif opt == telnetlib.NAWS:
            socket.sendall(telnetlib.IAC + telnetlib.WILL + opt)
            socket.sendall(telnetlib.IAC + telnetlib.SB
                           + telnetlib.NAWS
                           + struct.pack('!HH', width, height)
                           + telnetlib.IAC + telnetlib.SE)
        else:
            socket.sendall(telnetlib.IAC + telnetlib.WONT + opt[0])
    elif cmd == telnetlib.SB:
        if opt[0] == telnetlib.TTYPE and opt[1] == SEND:
            socket.sendall(telnetlib.IAC + telnetlib.SB
                           + telnetlib.TTYPE + IS + env_term
                           + chr(0) + telnetlib.IAC + telnetlib.SE)



"""
x84net message network web host for x/84.

This is a RESTful API server. Data is represented as JSON.

To configure the message network server, you must first configure the
web server. See x84/webserve.py for more information.

Add 'msgserve' to your 'modules' in the [web] section of default.ini.

Then, you must create a 'tag', that is, name your message network(s),
with a comma-delimited list of ``server_tags`` of the ``[msg]``
section.  The name of this tag will become the prefix of a series of
database files, and must be named appropriately (that is, contain
only alphanumerics).

Example default.ini settings:

[msg]
# The name of the message networks hosted
server_tags = x84net
"""
import logging
import hashlib
import json
import time
import web

#: response for general errors
RESP_FAIL = json.dumps({u'response': False, u'message': u'Error'})

#: token validation time in seconds
AUTH_EXPIREY = 15

#: maximum number of messages to reply in batches
BATCH_MSGS = 20

#: primary json fields
VALIDATE_FIELDS = ('network', 'action', 'auth',)

#: sub-fields of 'message'
VALIDATE_MSG_KEYS = (u'author', u'recipient', u'subject',
                     u'parent', u'tags', u'ctime', u'body', )


def parse_auth(request_data):
    """
    Parse and return tuple of 'auth' token if valid.

    :raises ValueError: ``when`` value of token is out of bounds,
                        or could not be coerced to an integer.
    """
    values = request_data.get('auth', '').split('|')
    if len(values) != 3:
        raise ValueError('Token must be 3-part pipe-delimited')

    _, _, when = values
    when = int(when)
    if time.time() > when + AUTH_EXPIREY:
        raise ValueError('Token is from the future')
    elif time.time() - when > AUTH_EXPIREY:
        raise ValueError('Token too far in the past')

    return values


class MessageApi(object):

    """ message network web API endpoint """

    # todo: validate messages much earlier (here, not below the scissor-line)

    def GET(self, network, last):
        """ GET method - pull messages """
        log = logging.getLogger(__name__)
        if 'HTTP_AUTH_X84NET' not in web.ctx.env:
            raise server_error(
                log_func=log.info,
                log_msg=('request without header Auth-X84net.'),
                status_exc=web.NoMethod)

        # prepare request for message, last is the highest
        # index previously received by client
        response_data = get_response(request_data={
            'auth': web.ctx.env['HTTP_AUTH_X84NET'],
            'network': network,
            'action': 'pull',
            'last': max(0, int(last)),
        })

        # return response data as json (200 OK)
        return self._jsonify(response_data, log)

    def PUT(self, network, *_):
        """ PUT method - post messages. """
        log = logging.getLogger(__name__)
        if 'HTTP_AUTH_X84NET' not in web.ctx.env:
            raise server_error(
                log_func=log.info,
                log_msg='request without header Auth-X84net.',
                status_exc=web.NoMethod)

        # parse incoming message
        webdata = web.input()
        response_data = get_response(request_data={
            'auth': web.ctx.env['HTTP_AUTH_X84NET'],
            'network': network,
            'action': 'push',
            'message': json.loads(webdata.message)
        })

        # return response data as json
        return self._jsonify(response_data, log)

    @staticmethod
    def _jsonify(response_data, log):
        """
        Return ``response_data`` as json.

        :raises web.HTTPError: response_data failed to encode to json.
        """
        try:
            return json.dumps(response_data)
        except ValueError as err:
            log.error('{err}: response_data={response_data!r}'.format(
                err=err, response_data=response_data))
            raise web.HTTPError('500 Server Error', {}, RESP_FAIL)


def web_module():
    """
    Return dictionary of url routes and function mappings for this module.

    Called by x84/webserve.py on server start.
    """
    return {
        'urls': ('/messages/([^/]+)/([^/]*)/?', 'messages'),
        'funcs': {
            'messages': MessageApi
        }
    }

# Above is the module definition.
#
# ---8<---
#
# Below is the method for serving requests and some helper funcs.


def server_error(log_func, log_msg, status_exc):
    """
    Helper function for logging and returning web.HTTPError.

    :param callable log_func: logging function.
    :param str log_msg: log message.
    :param web.HTTPError status_exc: exception class
    :raises: status_exc instance
    :returns: does not return.
    """
    exc = status_exc()
    log_func('{0}: {1}'.format(exc, log_msg))
    raise exc


def serve_messages_for(board_id, request_data, db_source):
    """ Reply-to api client request to receive new messages. """
    # pylint: disable=R0914
    #         Too many local variables (16/15)
    from x84.bbs import DBProxy, msgbase
    from x84.bbs.msgbase import to_utctime
    log = logging.getLogger(__name__)
    # log.error(msg)
    db_tags = DBProxy(msgbase.TAGDB, use_session=False)
    db_messages = DBProxy(msgbase.MSGDB, use_session=False)

    def message_owned_by(msg_id, board_id):
        """ Whether given message is owned by specified board. """
        return (msg_id in db_source and
                db_source[msg_id] == board_id)

    def msgs_after(idx=None):
        """
        Generator of network messages following index ``idx```.

        If ``idx`` is None, all messages are returned.
        """
        for msg_id in db_tags.get(request_data['network'], []):
            if idx is None:
                yield db_messages[idx]
            elif (int(msg_id) > int(idx) and
                  not message_owned_by(msg_id, board_id)):
                yield db_messages[msg_id]

    last_seen = request_data.get('last', None)
    pending_messages = msgs_after(last_seen)
    return_messages = list()
    num_sent = 0
    for num_sent, msg in enumerate(pending_messages, start=1):
        return_messages.append({
            u'id': msg.idx,
            u'author': msg.author,
            u'recipient': msg.recipient,
            u'parent': msg.parent,
            u'subject': msg.subject,
            u'tags': list(msg.tags ^ set([request_data['network']])),
            u'ctime': to_utctime(msg.ctime),
            u'body': msg.body
        })
        if num_sent >= BATCH_MSGS:
            log.warn('[{request_data[network]}] Batch limit reached for '
                     'board {board_id}; halting'
                     .format(request_data=request_data, board_id=board_id))
            break

    if num_sent > 0:
        log.info('[{request_data[network]}] {num_sent} messages '
                 'served to {board_id}'.format(request_data=request_data,
                                               num_sent=num_sent,
                                               board_id=board_id))

    return {u'response': True, u'messages': return_messages}


def receive_message_from(board_id, request_data,
                         db_source, db_transactions):
    """ Reply-to api client request to post a new message. """
    from x84.bbs.msgbase import to_localtime, Msg
    log = logging.getLogger(__name__)

    if 'message' not in request_data:
        raise server_error(
            log_func=log.info,
            log_msg="request data missing 'message' content",
            status_exc=web.BadRequest)

    pullmsg = request_data['message']

    # validate
    for key in (_key for _key in VALIDATE_MSG_KEYS if _key not in pullmsg):
        raise server_error(
            log_func=log.info,
            log_msg=("request data 'message' missing sub-field {key!r}"
                     .format(key=key)),
            status_exc=web.BadRequest)

    msg = Msg()
    msg.author = pullmsg['author']
    msg.recipient = pullmsg['recipient']
    msg.subject = pullmsg['subject']
    msg.parent = pullmsg['parent']
    msg.tags = set(pullmsg['tags'] + [request_data['network']])
    msg.body = pullmsg['body']

    # ?? is this removing millesconds, or ?
    _ctime = to_localtime(pullmsg['ctime'].split('.', 1)[0])

    msg.save(send_net=False, ctime=_ctime)
    with db_source, db_transactions:
        db_source[msg.idx] = board_id
        db_transactions[msg.idx] = msg.idx

    web.ctx.status = '201 Created'
    return {u'response': True, u'id': msg.idx}


def get_response(request_data):
    """ Serve one API server request and return. """
    # pylint: disable=R0914
    #         Too many local variables (16/15)
    from x84.bbs import DBProxy, get_ini
    log = logging.getLogger(__name__)

    # validate primary json request keys
    for key in (_key for _key in VALIDATE_FIELDS
                if _key not in request_data):
        raise server_error(
            log_func=log.info,
            log_msg=("request data 'message' missing sub-field {key!r}"
                     .format(key=key)),
            status_exc=web.BadRequest)

    # validate this server offers such message network
    server_tags = get_ini(section='msg', key='server_tags', split=True)

    if not request_data['network'] in server_tags:
        raise server_error(
            log_func=log.warn,
            log_msg=('[{data[network]}] not in server_tags ({server_tags})'
                     .format(data=request_data, server_tags=server_tags)),
            status_exc=web.NotFound)

    tag = request_data['network']

    # validate authentication token
    try:
        board_id, token, auth_tmval = parse_auth(request_data)
    except ValueError as err:
        raise server_error(
            log_func=log.warn,
            log_msg=('[{data[network]}] Bad token: {err}'
                     .format(data=request_data, err=err)),
            status_exc=web.Unauthorized)

    else:
        log.debug('[{data[network]}] board_id={board_id} request'
                  ': {data[action]}'.format(data=request_data,
                                            board_id=board_id))

    # validate board auth-key
    keysdb = DBProxy('{0}keys'.format(tag), use_session=False)
    try:
        client_key = keysdb[board_id]
    except KeyError:
        raise server_error(
            log_func=log.warn,
            log_msg=('[{data[network]}] board_id={board_id}'
                     ': No such key for this network'
                     .format(data=request_data,
                             board_id=board_id)),
            status_exc=web.Unauthorized)
    else:
        server_key = hashlib.sha256('{0}{1}'.format(client_key, auth_tmval))
        if token != server_key.hexdigest():
            raise server_error(
                log_func=log.warn,
                log_msg=('[{data[network]}] board_id={board_id}'
                         ': auth-key mismatch'
                         .format(data=request_data,
                                 board_id=board_id)),
                status_exc=web.Unauthorized,  # Unauthorized
            )

    # these need to be better named for their transmission direction,
    # its very clear how they are consumed as they are currently named.
    db_source = DBProxy('{0}source'.format(tag), use_session=False)
    db_transactions = DBProxy('{0}trans'.format(tag), use_session=False)

    if request_data.get('action', None) == 'pull':
        # client is requesting to pull messages
        return serve_messages_for(board_id=board_id,
                                  request_data=request_data,
                                  db_source=db_source)

    elif request_data.get('action', None) == 'push':
        # client is sending a message to the network
        return receive_message_from(board_id=board_id,
                                    request_data=request_data,
                                    db_source=db_source,
                                    db_transactions=db_transactions)

    raise server_error(
        log_func=log.info,
        log_msg=('[{data[network]}] Unknown action, {data[action]!r}'
                 .format(data=request_data)),
        status_exc=web.NoMethod)



""" Userbase record database and utility functions for x/84. """
import logging
from x84.bbs.dbproxy import DBProxy

FN_PASSWORD_DIGEST = None
GROUPDB = 'groupbase'
USERDB = 'userbase'


def list_users():
    """
    Returns all user handles.

    :rtype: list
    :returns list of user handles.
    """
    return [handle.decode('utf8')
            for handle in DBProxy(USERDB).keys()]


def get_user(handle):
    """
    Returns User record by handle.

    :rtype: User
    :returns: instance of :class:`User`
    """
    return DBProxy(USERDB)[handle]


def find_user(handle):
    """
    Discover and return matching user by ``handle``, case-insensitive.

    :returns: matching handle as str, or None if not found.
    :rtype: None or str.
    """
    for key in DBProxy(USERDB).keys():
        if handle.lower() == key.decode('utf8').lower():
            return key


class Group(object):

    """ A simple group record object. """

    def __init__(self, name, members=()):
        """ Class initializer. """
        self._name = name
        self._members = set(members)

    @property
    def name(self):
        """ Name of this group. """
        return self._name

    @name.setter
    def name(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._name = value

    @property
    def members(self):
        """ Members of this group as user handles. """
        return self._members

    def add(self, handle):
        """ Add user to group. """
        log = logging.getLogger(__name__)
        log.info("Group({!r}).add({!r})".format(self.name, handle))
        self._members.add(handle)

    def remove(self, handle):
        """ Remove user from group. """
        log = logging.getLogger(__name__)
        log.info("Group({!r}).remove({!r})".format(self.name, handle))
        self._members.remove(handle)

    def save(self):
        """ Save group record to database. """
        DBProxy(GROUPDB)[self.name] = self

    def delete(self):
        """ Delete group record, enforces referential integrity with Users. """
        udb = DBProxy(USERDB)
        for chk_user in self.members:
            user = udb[chk_user]
            if self.name in user.groups:
                user.group_del(self.name)
                user.save()
        del DBProxy(GROUPDB)[self.name]


class User(object):

    """ A simple user record. """

    def __init__(self, handle=u'anonymous'):
        """ Class initializer. """
        self._handle = handle
        self._password = (None, None)
        self._location = u''
        self._email = u''
        self._groups = set()
        self._calls = 0
        self._lastcall = 0

    @property
    def handle(self):
        """ User handle, also the database key. """
        return self._handle

    @handle.setter
    def handle(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._handle = value

    @property
    def password(self):
        """
        Password in encrypted form as tuple (salt, hash).

        Not generally used directly, but by :meth:`auth`.

        The ``setter`` of this property is provided a password
        in plain-text and encrypts it as given.

        If a password has not yet been set, it is (None, None).
        """
        return self._password

    @password.setter
    def password(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        log = logging.getLogger(__name__)
        from x84.bbs import ini
        if ini.CFG.getboolean('system', 'pass_ucase'):
            # facebook and mystic storage style, i wouldn't
            # recommend it though.
            self._password = get_digestpw()(value.upper())
        else:
            self._password = get_digestpw()(value)
        log.info("set password for user {!r}.".format(self.handle))

    def auth(self, try_pass):
        """
        Authenticate user with given password, ``try_pass``.

        :rtype: bool
        :returns: whether the password is correct.
        """
        from x84.bbs import ini
        pass_ucase = ini.CFG.getboolean('system', 'pass_ucase')
        assert isinstance(try_pass, unicode)
        assert len(try_pass) > 0
        assert self.password != (None, None), ('account is without password')
        salt = self.password[0]
        digestpw = get_digestpw()
        return (self.password == digestpw(try_pass, salt) or pass_ucase
                and self.password == digestpw(try_pass.upper(), salt))

    def __setitem__(self, key, value):
        # pylint: disable=C0111,
        #        Missing docstring
        log = logging.getLogger(__name__)
        adb = DBProxy(USERDB, 'attrs')

        if self.handle == 'anonymous':
            log.debug("set attr {!r} not possible for 'anonymous'".format(key))
            return

        with adb:
            if self.handle not in adb:
                adb[self.handle] = dict([(key, value), ])
            else:
                attrs = adb[self.handle]
                attrs.__setitem__(key, value)
                adb[self.handle] = attrs
        log.debug("set attr {!r} for user {!r}.".format(key, self.handle))
    __setitem__.__doc__ = dict.__setitem__.__doc__

    def get(self, key, default=None):
        # pylint: disable=C0111,
        #        Missing docstring
        from x84.bbs import ini
        log = logging.getLogger(__name__)
        adb = DBProxy(USERDB, 'attrs')

        if self.handle not in adb:
            if ini.CFG.getboolean('session', 'tap_db'):
                log.debug('User({!r}).get(key={!r}) returns default={!r}'
                          .format(self.handle, key, default))
            return default

        attrs = adb.get(self.handle, {})
        if key not in attrs:
            if ini.CFG.getboolean('session', 'tap_db'):
                log.debug('User({!r}.get(key={!r}) returns default={!r}'
                          .format(self.handle, key, default))
            return default

        if ini.CFG.getboolean('session', 'tap_db'):
            log.debug('User({!r}.get(key={!r}) returns value.'
                      .format(self.handle, key))
        return attrs[key]
    get.__doc__ = dict.get.__doc__

    def __getitem__(self, key):
        # pylint: disable=C0111,
        #        Missing docstring
        return DBProxy(USERDB, 'attrs')[self.handle][key]
    __getitem__.__doc__ = dict.__getitem__.__doc__

    def __delitem__(self, key):
        # pylint: disable=C0111,
        #        Missing docstring
        log = logging.getLogger(__name__)
        uadb = DBProxy(USERDB, 'attrs')
        with uadb:
            # retrieve attributes from uadb,
            attrs = uadb.get(self.handle, {})
            # delete attribute if exists
            if key in attrs:
                attrs.__delitem__(key)
                uadb[self.handle] = attrs
                log.info("User({!r}) delete attr {!r}."
                         .format(self.handle, key))
    __delitem__.__doc__ = dict.__delitem__.__doc__

    @property
    def groups(self):
        """ Set of groups user is a member of (set of strings). """
        return self._groups

    def group_add(self, group):
        """ Add user to group. """
        return self._groups.add(group)

    def group_del(self, group):
        """ Remove user from group. """
        return self._groups.remove(group)

    def save(self):
        """ Save user record to database. """
        log = logging.getLogger(__name__)
        assert isinstance(self._handle, unicode), ('handle must be unicode')
        assert len(self._handle) > 0, ('handle must be non-zero length')
        assert (None, None) != self._password, ('password must be set')
        assert self._handle != u'anonymous', ('anonymous may not be saved.')
        udb = DBProxy(USERDB)
        with udb:
            if 0 == len(udb) and self.is_sysop is False:
                log.warn('{!r}: First new user becomes sysop.'
                         .format(self.handle))
                self.group_add(u'sysop')
            is_new = self.handle not in udb
            udb[self.handle] = self
            if is_new:
                log.info("saved new user '%s'.", self.handle)
        adb = DBProxy(USERDB, 'attrs')
        with adb:
            if self.handle not in adb:
                adb[self.handle] = dict()
        self._apply_groups()

    def delete(self):
        """ Remove user from user and group databases. """
        log = logging.getLogger(__name__)
        gdb = DBProxy(GROUPDB)
        with gdb:
            for gname in self._groups:
                group = gdb[gname]
                if self.handle in group.members:
                    group.remove(self.handle)
                    group.save()
        udb = DBProxy(USERDB)
        with udb:
            del udb[self.handle]
        log.info("deleted user '%s'.", self.handle)

    @property
    def is_sysop(self):
        """ Whether the user is in the 'sysop' group. """
        return u'sysop' in self._groups

    @property
    def lastcall(self):
        """ Time last called, ``time.time()`` epoch-formatted (float). """
        return self._lastcall

    @lastcall.setter
    def lastcall(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._lastcall = value

    @property
    def calls(self):
        """Legacy, number of times user has 'called' this board."""
        return self._calls

    @calls.setter
    def calls(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._calls = value

    @property
    def location(self):
        """ Legacy, used as a geographical location, group names, etc. """
        return self._location

    @location.setter
    def location(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._location = value

    @property
    def email(self):
        """ E-mail address. May be used for password resets. """
        return self._email

    @email.setter
    def email(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._email = value

    def _apply_groups(self):
        """ Enforce referential integrity of user's groups. """
        log = logging.getLogger(__name__)
        gdb = DBProxy(GROUPDB)
        with gdb:
            for chk_grp in self._groups:
                if chk_grp not in gdb:
                    gdb[chk_grp] = Group(chk_grp, set([self.handle]))
                    log.info("created group {!r} for user {!r}."
                             .format(chk_grp, self.handle))
                # ensure membership in existing groups
                group = gdb[chk_grp]
                if self.handle not in group.members:
                    group.add(self.handle)
                    group.save()
            for gname, group in gdb.items():
                if gname not in self._groups and self.handle in group.members:
                    group.remove(self.handle)
                    group.save()


def _digestpw_bcrypt(password, salt=None):
    """ Password digest using bcrypt (optional-preferred). """
    import bcrypt
    if not salt:
        salt = bcrypt.gensalt()
    if isinstance(password, unicode):
        password = password.encode('utf8')
    return salt, bcrypt.hashpw(password, salt)


def _digestpw_internal(password, salt=None):
    """ Password digest using regular python libs (slow). """
    import hashlib
    import base64
    import os
    if not salt:
        salt = base64.b64encode(os.urandom(32))
    digest = salt + password
    for _ in range(0, 100000):
        # pylint: disable=E1101
        #         Module 'hashlib' has no 'sha256'
        digest = hashlib.sha256(digest).hexdigest()
    return salt, digest


def _digestpw_plaintext(password, salt=None):
    """ No password digest, just store the passwords in plain text. """
    if not salt:
        salt = 'none'
    return salt, password


def get_digestpw():
    """ Returns singleton to password digest routine. """
    global FN_PASSWORD_DIGEST
    if FN_PASSWORD_DIGEST is not None:
        return FN_PASSWORD_DIGEST

    from x84.bbs.ini import get_ini
    FN_PASSWORD_DIGEST = {
        'bcrypt': _digestpw_bcrypt,
        'internal': _digestpw_internal,
        'plaintext': _digestpw_plaintext,
    }.get(get_ini('system', 'password_digest'))
    return FN_PASSWORD_DIGEST


def check_new_user(username):
    """ Boolean return when username matches ``newcmds`` ini cfg. """
    from x84.bbs import get_ini
    matching = get_ini(section='matrix',
                       key='newcmds',
                       split=True)
    allowed = get_ini(section='nua',
                      key='allow_apply',
                      getter='getboolean')
    return allowed and username in matching


def check_bye_user(username):
    """ Boolean return when username matches ``byecmds`` in ini cfg. """
    from x84.bbs import get_ini
    matching = get_ini(section='matrix', key='byecmds', split=True)
    return matching and username in matching


def check_anonymous_user(username):
    """ Boolean return when user is anonymous and is allowed. """
    from x84.bbs import get_ini
    matching = get_ini(section='matrix',
                       key='anoncmds',
                       split=True)
    allowed = get_ini(section='matrix',
                      key='enable_anonymous',
                      getter='getboolean',
                      split=False)
    return allowed and username in matching


def check_user_password(username, password):
    """ Boolean return when username and password match user record. """
    from x84.bbs import find_user, get_user
    handle = find_user(username)
    if handle is None:
        return False
    user = get_user(handle)
    if user is None:
        return False
    return password and user.auth(password)


def parse_public_key(user_pubkey):
    """ Return paramiko key class instance of a user's public key text. """
    import paramiko

    if len(user_pubkey.split()) == 3:
        key_msg, key_data, _ = user_pubkey.split()
    elif len(user_pubkey.split()) == 2:
        key_msg, key_data = user_pubkey.split()
    elif len(user_pubkey.split()) == 1:
        # when no key-type is specified, assume rsa
        key_msg, key_data = 'ssh-rsa', user_pubkey
    else:
        raise ValueError('Malformed public key format: {0!r}'
                         .format(user_pubkey))
    try:
        key_bytes = key_data.decode('ascii')
    except UnicodeDecodeError:
        raise ValueError('Malformed public key encoding: {0!r}'
                         .format(key_data))
    decoded_keybytes = paramiko.py3compat.decodebytes(key_bytes)
    try:
        return {'ssh-rsa': paramiko.RSAKey,
                'ssh-dss': paramiko.DSSKey,
                'ecdsa-sha2-nistp256': paramiko.ECDSAKey,
                }.get(key_msg)(data=decoded_keybytes)
    except KeyError:
        raise ValueError('Malformed public key_msg: {0!r}'
                         .format(key_msg))


def check_user_pubkey(username, public_key):
    """ Boolean return when public_key matches user record. """
    from x84.bbs import find_user, get_user
    log = logging.getLogger(__name__)
    handle = find_user(username)
    if handle is None:
        return False
    user_pubkey = get_user(handle).get('pubkey', False)
    if not user_pubkey:
        log.debug('pubkey authentication by {0!r} but no '
                  'public key on record for the user.'
                  .format(username))
        return False
    try:
        stored_pubkey = parse_public_key(user_pubkey)
    except (ValueError, Exception):
        import sys
        (exc_type, exc_value, _) = sys.exc_info()
        log.debug('{0} for stored public key of user {1!r}: '
                  '{2}'.format(exc_type, username, exc_value))
    else:
        return stored_pubkey == public_key



"""
Sysop area script for x/84.

Currently, this only serves the purpose of adding new message networks.
"""

from x84.bbs import getsession, getterminal, echo, get_ini, DBProxy, LineEditor


MSG_NO_SERVER_TAGS = "no `server_tags' defined in ini file, section [msg]."


def view_leaf_msgnet(server_tag=None, board_id=None):
    if server_tag is None:
        server_tags = get_ini(section='msg', key='server_tags', split=True)
        if not server_tags:
            raise ValueError("no `server_tags' defined in ini file, "
                             "section [msg].")
        # RECURSION
        for _st in server_tags:
            view_leaf_msgnet(server_tag=_st, board_id=None)
        return

    if board_id is None:
        board_ids = DBProxy('{0}keys'.format(server_tag)).keys()
        for _bid in board_ids:
            # RECURSION
            view_leaf_msgnet(server_tag=server_tag, board_id=_bid)
        return

    with DBProxy('{0}keys'.format(server_tag)) as key_db:
        echo(u'\r\n[msgnet_{0}]'.format(server_tag))
        echo(u'\r\nurl_base = https://{addr}:{port}/'
             .format(addr=get_ini('web', 'addr'),
                     port=get_ini('web', 'port')))
        echo(u'\r\nboard_id = {0}'.format(board_id))
        echo(u'\r\ntoken = {0}'.format(key_db[board_id]))
        echo(u'\r\npoll_interval = 300')
        echo(u'\r\n')
        echo(u'\r\n[msg]')
        echo(u'\r\nnetwork_tags = {0}'.format(server_tag))
    echo(u'\r\n')
    echo(u'-' * 40)


def add_leaf_msgnet():
    import cryptography.fernet
    server_tags = get_ini(section='msg', key='server_tags', split=True)
    if not server_tags:
        raise ValueError(MSG_NO_SERVER_TAGS)

    if len(server_tags) == 1:
        server_tag = server_tags[0]
    else:
        while True:
            echo('chose a server tag: ')
            idx = 0
            for idx, tag in server_tags:
                echo(u'\r\n{0}. {1}'.format(idx, tag))
            echo(u'\r\n: ')
            inp = LineEditor(width=len(str(idx)).read())
            if inp is None:
                return
            try:
                server_tag = server_tags[int(inp)]
                break
            except ValueError:
                pass

    with DBProxy('{0}keys'.format(server_tag)) as key_db:
        board_id = max(map(int, key_db.keys()) or [-1]) + 1
        client_key = cryptography.fernet.Fernet.generate_key()
        key_db[board_id] = client_key
    echo(u'\r\n')
    echo(u'-' * 40)
    view_leaf_msgnet(server_tag, board_id)


def main():
    session, term = getsession(), getterminal()
    assert session.user.is_sysop

    dirty = True
    while True:
        if dirty:
            echo(u'\r\n\r\nmessage network functions:\r\n')
            echo(u'    [a]dd new leaf node.\r\n')
            echo(u'    [v]iew leaf nodes.\r\n')
            echo(u'\r\n\r\n')
            echo(u'[q]uit\r\n')
            dirty = False
        echo(u'\r\nsysop cmd: ')
        inp = term.inkey()
        if inp.lower() == u'q':
            echo(inp)
            echo(u'\r\n')
            return
        elif inp.lower() == u'a':
            echo(inp)
            add_leaf_msgnet()
            dirty = True
        elif inp.lower() == u'v':
            echo(inp)
            echo(u'\r\n')
            echo(u'-' * 40)
            view_leaf_msgnet()
            echo(u'\r\n\r\nPress any key.')
            term.inkey()
            dirty = True



""" Who's online script for x/84. """
import time
POLL_KEY = 0.25  # blocking ;; how often to poll keyboard
POLL_INF = 2.00  # seconds elapsed until re-ask clients for more details
POLL_AYT = 4.00  # seconds elapsed until global 'are you there?' is checked,
POLL_OUT = 0.50  # seconds elapsed before screen updates


def request_info(sid):
    """ Send info-req event to target session id ``sid``. """
    from x84.bbs import getsession
    session = getsession()
    session.send_event('route', (sid, 'info-req', session.sid,))


def banner():
    """ Returns string suitable for displaying banner """
    from x84.bbs import getterminal, showart
    import os
    term = getterminal()
    banner = '\r\n'
    for line in showart(
            os.path.join(os.path.dirname(__file__), 'art', 'online.ans'), 'cp437'):
        banner = banner + term.move_x(max(0, (term.width / 2) - 40)) + line
    return (banner)


def describe(sessions):
    """
    Returns unicode string suitable for describing the activity of
    session id's of array ``sessions``.
    """
    from x84.bbs import getsession, getterminal, ini
    slen = lambda sessions: len(u'%d' % (len(sessions),))
    session, term = getsession(), getterminal()
    max_user = ini.CFG.getint('nua', 'max_user')

    text = u'\r\n'.join(([u''.join((
        term.move_x(max(0, (term.width / 2) - 40)), term.green,
        u'%*d' % (5 + slen(sessions), node), u' ' * 7, term.normal,
        u'%4is' % (attrs.get('idle', 0),), u' ', u' ' * 8,
        (term.bold_red(u'%-*s' % (max_user, (
            u'** diSCONNECtEd' if 'delete' in attrs
            else attrs.get('handle', u'** CONNECtiNG')),)
        ) if attrs.get('handle', u'') != session.user.handle
            else term.red(u'%-*s' % (max_user, session.user.handle))),
        term.green(u'       '),
        term.yellow((attrs.get('activity', u''))
                    if attrs.get('sid') != session.sid else
                    term.yellow(session.activity)),
    )) for node, (_, attrs) in get_nodes(sessions)]))

    return text + '\r\n'


def get_nodes(sessions):
    """ Given an array of sessions, assign an arbitrary 'node' number """
    return enumerate(sorted(sessions.items()))


def heading():
    """ Return string suitable for display heading. """
    from x84.bbs import getterminal, showart
    import os
    term = getterminal()
    bar = ''
    for line in showart(
            os.path.join(os.path.dirname(__file__), 'art', 'onlinebar.ans'), 'topaz'):
        bar = bar + term.move_x(max(0, (term.width / 2) - 40)) + line
    return u'\r\n'.join((
        u'\r\n'.join([term.center(pline, (term.width))
                      for pline in prompt()]),
        u'\r\n', bar))


def prompt():
    """
    Return string suitable for displaying prompt and available commands.
    """
    from x84.bbs import getsession, getterminal
    session, term = getsession(), getterminal()
    decorate = lambda key, desc: u''.join((
        u'(', term.magenta_underline(key,),
        u')', term.cyan(desc.split()[0]), u' ',
        u' '.join(desc.split()[1:]), u' ',))
    return term.wrap(u''.join((
        u' ' * 2,
        term.green_reverse(':keys'), u' ',
        decorate('c', 'hAt USR'),
        decorate('s', 'ENd MSG'),
        (u''.join((
            decorate('d', 'iSCONNECt SiD'),
            decorate('e', 'diT USR'),
            decorate('v', 'iEW SiD AttRS'),
            u' ',)) if 'sysop' in session.user.groups else u''),
        decorate('Escape/q', 'Uit'),
        decorate('Spacebar', 'REfRESh'),
    )), int(term.width * .8), subsequent_indent=u' ' * 8)


def get_node(sessions):
    """ Prompt user for session node, Returns node & session attributes. """
    from x84.bbs import ini, LineEditor, echo
    max_user = ini.CFG.getint('nua', 'max_user')
    invalid = u'\r\ninvalid.'
    echo(u'\r\n\r\nNOdE: ')
    node = LineEditor(max_user).read()

    if node is None or 0 == len(node):
        # cancel
        return (None, None)

    try:
        node = int(node)
    except ValueError:
        # not an int
        echo(invalid)
        return (None, None)

    for tgt_node, (_, attrs) in get_nodes(sessions):
        if tgt_node == node:
            return (tgt_node, attrs)

    # not found
    echo(invalid)
    return (None, None)


def edit(sessions):
    """ Prompt for node and gosub profile.py script for user of target session.
    """
    from x84.bbs import gosub
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        gosub('profile', handle=tgt_session['handle'])
        return True


def chat(sessions):
    """
    Prompt for node and page target session for chat.
    Sysop will send session id of -1, indicating the chat is forced.
    """
    from x84.bbs import gosub, getsession
    (_, tgt_session) = get_node(sessions)
    if tgt_session and tgt_session['sid'] != getsession().sid:
        gosub('chat', dial=tgt_session['handle'],
              other_sid=tgt_session['sid'])


def view(sessions):
    """
    Prompt for node and view session details of target session.
    """
    from x84.bbs import echo, getterminal
    term = getterminal()
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        maxlen = max([len(key) for key in tgt_session.keys()])
        echo(u''.join((
            u'\r\n\r\n',
            u'\r\n'.join(['%s%s %s' % (
                term.bold('%*s' % (maxlen, key)),
                term.bold_green(':'),
                term.green(str(value)),)
                for key, value in sorted(tgt_session.items())]),
        )))
        return True


def disconnect(sessions):
    """
    Prompt for node and disconnect target session.
    """
    from x84.bbs import getsession
    session = getsession()
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        session.send_event('remote-disconnect', tgt_session['sid'])
        return True


def sendmsg(sessions):
    """
    Prompt for node and gosub 'writemsg' with recipient set to target user.
    """
    from x84.bbs import gosub, Msg
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        msg = Msg()
        msg.recipient = tgt_session['handle']
        msg.tags.add('private')
        gosub('writemsg', msg)
        return True


def main():
    """ Main procedure. """
    # pylint: disable=R0912,R0914,R0915
    #         Too many branches
    #         Too many local variables
    #         Too many statements
    from x84.bbs import getsession, getterminal, echo
    session, term = getsession(), getterminal()
    SELF_ID = session.sid
    ayt_lastfresh = 0

    def broadcast_ayt(last_update):
        """ Globally boradcast 'are-you-there' request. """
        if time.time() - last_update > POLL_AYT:
            session.send_event('global', ('AYT', session.sid,))
            last_update = time.time()
        return last_update

    sessions = dict()
    dirty = time.time()
    cur_row = 0

    while True:
        ayt_lastfresh = broadcast_ayt(ayt_lastfresh)
        inp = term.inkey(POLL_KEY)
        if session.poll_event('refresh') or (inp in (u' ', unichr(12))):
            dirty = time.time()
            cur_row = 0
        elif inp.lower() in (u'q', unichr(27)) or inp.code == term.KEY_EXIT:
            echo(u'\r\n\r\n')
            return
        elif inp.lower() == u'c':
            cur_row = 0 if chat(sessions) else cur_row
            dirty = time.time()
        elif inp.lower() == u's':
            cur_row = 0 if sendmsg(sessions) else cur_row
            dirty = time.time()
        elif inp and 'sysop' in session.user.groups:
            if inp.lower() == u'e':
                cur_row = 0 if edit(sessions) else cur_row
                dirty = time.time()
            elif inp.lower() == u'v':
                cur_row = 0 if view(sessions) else cur_row + 3
                dirty = time.time()
            elif inp.lower() == u'd':
                disconnect(sessions)
                dirty = time.time()

        # add sessions that respond to AYT
        data = session.poll_event('ACK')
        if data is not None:
            sid, handle = data
            if sid in sessions:
                sessions[sid]['handle'] = handle
            else:
                sessions[sid] = dict((
                    ('handle', handle),
                    ('lastfresh', time.time()),
                    ('lastasked', time.time()),))
                dirty = time.time()
                echo(u'\a')

        # update sessions that respond to info-req
        data = session.poll_event('info-ack')
        if data is not None:
            sid, attrs = data
            if sessions.get(sid, dict()).get('activity') != attrs['activity']:
                # and refresh screen if activity changes
                dirty = time.time()
            sessions[sid] = attrs
            sessions[sid]['lastfresh'] = time.time()

        # update our own session
        sessions[SELF_ID] = session.to_dict()
        sessions[SELF_ID]['lastfresh'] = time.time()

        # request that all sessions update if more stale than POLL_INF,
        # or is missing session info (only AYT replied so far!),
        # or has been displayed as 'Disconnected' (marked for deletion)
        for sid, attrs in sessions.items():
            if sid == SELF_ID:
                continue
            if attrs.get('idle', -1) == -1 or (
                    time.time() - attrs.get('lastfresh', 0) > POLL_INF
                    and time.time() - attrs.get('lastasked', 0) > POLL_INF):
                request_info(sid)
                attrs['lastasked'] = time.time()

        # prune users who haven't responded to AYT
        for sid, attrs in sessions.items():
            if time.time() - attrs['lastfresh'] > (POLL_AYT * 2):
                sessions[sid]['delete'] = 1
                dirty = time.time()

        if dirty is not None and time.time() - dirty > POLL_OUT:
            session.activity = u"Who's Online"
            otxt = describe(sessions)
            olen = len(otxt.splitlines())
            if 0 == cur_row or (cur_row + olen) >= term.height:
                otxt_b = banner()
                otxt_h = heading()
                cur_row = len(otxt_b.splitlines()) + len(otxt_h.splitlines())
                echo(u''.join((otxt_b, '\r\n', otxt_h, u'\r\n', otxt)))
            else:
                echo(u''.join((
                    u'\r\n',
                    '-'.center(term.width).rstrip(),
                    u'\r\n')))
                echo(otxt)
            cur_row += olen
            dirty = None

        # delete disconnected sessions
        for sid, attrs in sessions.items()[:]:
            if attrs.get('delete', 0) == 1:
                del sessions[sid]



""" Pager package for x/84. """
from x84.bbs.ansiwin import AnsiWindow
from x84.bbs.output import encode_pipe, decode_pipe
from x84.bbs.session import getterminal
from x84.bbs.output import echo

VI_KEYSET = {
    'refresh': [unichr(12), ],
    'home': [u'0'],
    'end': [u'G'],
    'up': [u'k', u'K'],
    'down': [u'j', u'J', u'\r'],
    'pgup': [u'b', u'B', u''],
    'pgdown': [u'f', u'F', u''],
    'exit': [u'q', u'Q', unichr(27), ],
}


class Pager(AnsiWindow):

    """ Scrolling viewer. """

    def __init__(self, *args, **kwargs):
        """
        Class initializer.

        :param int width: width of window.
        :param int height: height of window.
        :param int yloc: y-location of window.
        :param int xloc: x-location of window.
        :param str content: initial pager contents.
        :param dict colors: color theme.
        :param dict glyphs: bordering window character glyphs.
        :param dict keyset: command keys, global ``VI_KEYSET`` is default.
        """
        self._quit = False
        self.init_keystrokes(keyset=kwargs.pop('keyset', VI_KEYSET.copy()))
        _content = kwargs.pop('content', u'') or u''
        AnsiWindow.__init__(self, *args, **kwargs)
        self.content = _content
        self._position = self.position = kwargs.pop('position', 0) or 0

    def init_keystrokes(self, keyset):
        """ Sets keyboard keys for various editing keystrokes. """
        term = getterminal()
        self.keyset = keyset
        self.keyset['home'].append(term.KEY_HOME)
        self.keyset['end'].append(term.KEY_END)
        self.keyset['pgup'].append(term.KEY_PGUP)
        self.keyset['pgdown'].append(term.KEY_PGDOWN)
        self.keyset['up'].append(term.KEY_UP)
        self.keyset['down'].append(term.KEY_DOWN)
        self.keyset['down'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_ESCAPE)

    @property
    def quit(self):
        """ Whether a 'quit' character has been handled, such as escape. """
        return self._quit

    @property
    def position_last(self):
        """ Previous position before last move. """
        return self._position_last

    @property
    def position(self):
        """ Index of content buffer displayed at top of window. """
        return self._position

    @position.setter
    def position(self, pos):
        # pylint: disable=C0111
        #         Missing docstring
        self._position_last = self._position

        # assign and bounds check
        self._position = min(max(0, pos), self.bottom)
        self.moved = (self._position_last != self._position)

    @property
    def visible_content(self):
        """ Content that is visible in window. """
        return self._content[self.position:self.position + self.visible_height]

    @property
    def visible_bottom(self):
        """ Bottom-most window row that contains content. """
        if self.bottom < self.visible_height:
            return self.bottom
        return len(self.visible_content) - 1

    @property
    def bottom(self):
        """ Bottom-most position that contains content. """
        maximum = self.visible_height
        return max(0, maximum - self.visible_height)

    def process_keystroke(self, keystroke):
        """
        Process the keystroke and return string to refresh.

        :param blessed.keyboard.Keystroke keystroke: input from ``inkey()``.
        :rtype: str
        :returns: string sequence suitable for refresh.
        """
        self.moved = False
        rstr = u''
        if (keystroke in self.keyset['refresh'] or
                keystroke.code in self.keyset['refresh']):
            rstr += self.refresh()
        elif (keystroke in self.keyset['up'] or
              keystroke.code in self.keyset['up']):
            rstr += self.move_up()
        elif (keystroke in self.keyset['down'] or
              keystroke.code in self.keyset['down']):
            rstr += self.move_down()
        elif (keystroke in self.keyset['home'] or
              keystroke.code in self.keyset['home']):
            rstr += self.move_home()
        elif (keystroke in self.keyset['end'] or
              keystroke.code in self.keyset['end']):
            rstr += self.move_end()
        elif (keystroke in self.keyset['pgup'] or
              keystroke.code in self.keyset['pgup']):
            rstr += self.move_pgup()
        elif (keystroke in self.keyset['pgdown'] or
              keystroke.code in self.keyset['pgdown']):
            rstr += self.move_pgdown()
        elif (keystroke in self.keyset['exit'] or
              keystroke.code in self.keyset['exit']):
            self._quit = True
        return rstr

    def read(self):
        """
        Blocking read-eval-print loop for pager.

        Processes user input, taking action upon and refreshing pager
        until the escape key is pressed.

        :rtype: None
        """
        self._quit = False
        echo(self.refresh())
        term = getterminal()
        while not self.quit:
            echo(self.process_keystroke(term.inkey()))

    def move_home(self):
        """
        Scroll to top and return refresh string.

        :rtype: str
        """
        self.position = 0
        if self.moved:
            return self.refresh()
        return u''

    def move_end(self):
        """
        Scroll to bottom and return refresh string.

        :rtype: str
        """
        self.position = len(self._content) - self.visible_height
        if self.moved:
            return self.refresh()
        return u''

    def move_pgup(self, num=1):
        """
        Scroll up ``num`` pages and return refresh string.

        :rtype: str
        """
        self.position -= (num * (self.visible_height))
        return self.refresh() if self.moved else u''

    def move_pgdown(self, num=1):
        """
        Scroll down ``num`` pages and return refresh string.

        :rtype: str
        """
        self.position += (num * (self.visible_height))
        return self.refresh() if self.moved else u''

    def move_down(self, num=1):
        """
        Scroll down ``num`` rows and return refresh string.

        :rtype: str
        """
        self.position += num
        if self.moved:
            return self.refresh()
        return u''

    def move_up(self, num=1):
        """
        Scroll up ``num`` rows and return refresh string.

        :rtype: str
        """
        self.position -= num
        if self.moved:
            return self.refresh()
        return u''

    def refresh_row(self, row):
        """
        Return unicode string suitable for refreshing pager row.

        :param int row: target row by visible index.
        :rtype: str
        """
        term = getterminal()
        ucs = u''
        if row < len(self.visible_content):
            ucs = self.visible_content[row]
        disp_position = self.pos(row + self.ypadding, self.xpadding)
        return u''.join((term.normal,
                         disp_position,
                         self.align(ucs),
                         term.normal))

    def refresh(self, start_row=0):
        """
        Return unicode string suitable for refreshing pager window.

        :param int start_row: refresh from only visible row 'start_row'
                              and downward. This can be useful if only
                              the last line is modified; or in an
                              'insert' operation: only the last line
                              need be refreshed.
        :rtype: str
        """
        term = getterminal()
        return u''.join(
            [term.normal] + [
                self.refresh_row(row)
                for row in range(start_row, len(self.visible_content))
            ] + [term.normal])

    def update(self, ucs):
        """
        Update content buffer with newline-delimited text.

        :rtype: str
        """
        self.content = ucs
        return self.refresh()

    @property
    def content(self):
        """
        Content of pager.

        Return value is "pipe encoded" by :func:`encode_pipe`.
        :rtype: str
        """
        return encode_pipe('\r\n'.join(self._content))

    @content.setter
    def content(self, ucs_value):
        # pylint: disable=C0111
        #         Missing method docstring
        self._content = self._content_wrap(decode_pipe(ucs_value))

    def _content_wrap(self, ucs):
        """ Return word-wrapped text ``ucs`` that contains newlines. """
        term = getterminal()
        lines = []
        for line in ucs.splitlines():
            if line.strip():
                lines.extend(term.wrap(line, self.visible_width - 1))
            else:
                lines.append(u'')
        return lines

    def append(self, ucs):
        """
        Update content buffer with additional line(s) of text.

        "pipe codes" in ``ucs`` are decoded by :func:`decode_pipe`.

        :param str ucs: unicode string to append-to content buffer.
        :rtype str
        :return: terminal sequence suitable for refreshing window.
        """
        self._content.extend(self._content_wrap(decode_pipe(ucs)))
        return self.move_end() or self.refresh(self.bottom)



"""
SSH Matrix for x/84.

This script is the default session entry point for all ssh connections.

As our transport is ssh -- we've *already* authenticated the user,
though, when 'anonymous' login is enabled, or the handle is one of
the 'new user handles', we pass the mutually exclusive boolean keyword
arguments 'anonymous' or 'new'.

The argument 'username' is always set as the 'user@' argument of the
connecting ssh client -- it could be 'anonymous' or 'new', or any
case insensitive match of a user handle -- it does not necessarily
guarantee that the user exists!

When set, this is a user that should be found under find_user(username)
who may have already authenticated by some various means.
"""


def main(anonymous=False, new=False, username=''):
    """ Main procedure. """
    from x84.bbs import echo, goto, find_user, ini
    topscript = ini.CFG.get('matrix', 'topscript')
    nuascript = ini.CFG.get('nua', 'script')

    # http://www.termsys.demon.co.uk/vtansi.htm
    # disable line-wrapping
    echo(unichr(27) + u'[7l')

    # http://www.xfree86.org/4.5.0/ctlseqs.html
    # Save xterm icon and window title on stack.
    echo(unichr(27) + u'[22;0t')

    if anonymous:
        # user ssh'd in as anonymous@
        goto(topscript, 'anonymous')
    elif new:
        # user ssh'd in as new@
        goto(nuascript)

    handle = find_user(username)
    assert handle is not None, handle
    goto(topscript, handle=handle)



"""
Hacker news script for x/84.

Lets you browse and read popular articles on news.ycombinator.com.
"""
# std imports
from __future__ import division
import collections
import urlparse
import textwrap
import math
import sys

# local
from x84.bbs import getsession, getterminal, echo, LineEditor, get_ini

# 3rd-party
import feedparser
import html2text
import requests

#: fontset for SyncTerm emulator
SYNCTERM_FONT = get_ini(
    section='hackernews', key='syncterm_font'
) or 'topaz'

#: color of titlebars when viewing article summary
COLOR_MAIN = get_ini(
    section='hackernews', key='color_main'
) or 'black_on_red'

#: color of titlebars when viewing article
COLOR_VIEW = get_ini(
    section='hackernews', key='color_view'
) or "black_on_magenta"

USER_AGENT = 'Lynx/2.8.7rel.2 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/1.0.0a'
RSS_URL = 'https://news.ycombinator.com/rss'
RSS_TITLE = 'Hacker News'
ARTICLE_LIMIT = 100
REQUEST_TIMEOUT = 10

#: structure defines an article
Article = collections.namedtuple(
    'Article', ['title', 'link', 'comments', 'netloc'])

#: structure defines a key-movement by mapping to ``KEYSET``.
Direction = collections.namedtuple(
    'Direction', ['mapping', 'modifier', 'clamp'])

#: structure defines key movements for :func:`do_movement`
KEYSET = {
    'up': ['KEY_UP', u'k'],
    'sup': ['KEY_SUP', u'K'],
    'pgup': ['KEY_PGUP', u'b'],
    'home': ['KEY_HOME', u'0'],
    'down': ['KEY_DOWN', u'j'],
    'sdown': ['KEY_SDOWN', u'J'],
    'pgdown': ['KEY_PGDOWN', u'f', u' '],
    'end': ['KEY_END', u'G'],
}

MSG_NOTEXT = ("I'm sorry, the page you requested cannot be displayed in "
              "textmode.  It's probably not worth *reading*, anyway ...")


def get_keyset(term):
    """ Evaluate and return a keymap for given ``term``. """
    return {
        movement: [
            (val if not val.startswith('KEY')
             else getattr(term, val))
            for val in KEYSET[movement]
        ] for movement in KEYSET
    }


def do_movement(inp, scroll_idx, height, bottom, keyset):
    """ Return new scroll_idx if ``inp`` is a matching movement. """

    directions = (
        # up directions
        Direction(mapping='up',
                  modifier=-1,
                  clamp=lambda idx: max(0, idx)),
        Direction(mapping='sup',
                  modifier=-5,
                  clamp=lambda idx: max(0, idx)),
        Direction(mapping='pgup',
                  modifier=(height - 1) * -1,
                  clamp=lambda idx: max(0, idx)),
        Direction(mapping='home',
                  modifier=scroll_idx * -1,
                  clamp=lambda idx: max(0, idx)),

        # down directions
        Direction(mapping='down',
                  modifier=1,
                  clamp=lambda idx: min(bottom, idx)),
        Direction(mapping='sdown',
                  modifier=5,
                  clamp=lambda idx: min(bottom, idx)),
        Direction(mapping='pgdown',
                  modifier=height - 1,
                  clamp=lambda idx: min(bottom, idx)),
        Direction(mapping='end',
                  modifier=bottom - scroll_idx,
                  clamp=lambda idx: min(bottom, idx)),
    )

    # evaluate and return modified scroll_idx for any matching keyset.
    for direction in directions:
        if (inp.code in keyset[direction.mapping] or
                inp in keyset[direction.mapping]):
            return direction.clamp(scroll_idx + direction.modifier)
    return scroll_idx


def get_article(term, articles):
    """ Prompt for an article number, return matching article. """
    moveto_lastline = term.move(term.height, 0)
    width = term.width
    if term.kind.startswith('ansi'):
        # bah syncterm
        moveto_lastline = term.move(term.height - 1, 0)
        width -= 1
    echo(u''.join((
        moveto_lastline + getattr(term, COLOR_MAIN),
        term.center('', width),
        moveto_lastline,
    )))
    echo(u':: enter article #: ')
    article_idx = LineEditor(
        width=len(str(ARTICLE_LIMIT)),
        colors={'highlight': getattr(term, COLOR_MAIN)}
    ).read()
    if article_idx is None:
        # pressed escape
        return None
    try:
        return articles[int(article_idx) - 1]
    except (ValueError, IndexError):
        # not an integer, or out of range
        return None


def render_article(term, html_text):
    """ Render and return html text of article as text. """
    html_renderer = html2text.HTML2Text(bodywidth=term.width - 1)
    html_renderer.ignore_links = True
    html_renderer.ignore_images = True

    text_wrapped = []
    for line in html_renderer.handle(html_text).splitlines():
        if len(line) < term.width:
            text_wrapped.append(line)
        else:
            # html2text does not always honor `bodywidth',
            # calculate indentation (line up with previous indent)
            # and textwrap again.
            _subsq_indent = 0
            for _subsq_indent, char in enumerate(line):
                if not char.isspace():
                    break
            _indent = u' ' * _subsq_indent
            text_wrapped.extend(textwrap.wrap(line, term.width - 1,
                                              subsequent_indent=_indent))
    final = [_text.rstrip() for _text in text_wrapped]

    if not final or not any(_line for _line in final):
        # no text was rendered by html2text
        final = [''] * (term.height // 2)
        final.extend(textwrap.wrap(MSG_NOTEXT, term.width - 1))
    return final


def get_article_summaries(term, articles):
    """ Render list of articles summary. """
    results = []
    for idx, article in enumerate(articles, 1):
        _idx = '{0}{1}'.format(idx, term.bold_black('.'))
        _title = article.title
        _netloc = term.bold_black('({0})'.format(article.netloc))
        results.append(u' '.join((_idx, _title, _netloc)))
    return results


def render_articles_summary(term, scroll_idx, height, articles):
    """ Render and return articles summary in-view. """
    # _vpadd: larger terminals get an extra space between summaries
    _vpadd = 2 if term.height > 25 else 1
    _endline = term.clear_eol + u'\r\n'

    line_no = 0

    # render all articles,
    output = u''
    _start, _end = scroll_idx, (scroll_idx + (height // _vpadd))
    for line in articles[_start:_end]:
        # calculate indentation (after first ' ').
        _subsq_indent = ((term.strip_seqs(line[:30]).find(' ') or -1) + 1)

        # render each line, wrapping to terminal width
        for subline in term.wrap(line, subsequent_indent=u' ' * _subsq_indent):
            if line_no < height:
                output = u''.join((output, subline, _endline))
                line_no += 1

        # add additional padding between articles if room remains
        if line_no + (_vpadd - 1) < height:
            output = u''.join((output, _endline * (_vpadd - 1)))
            line_no += (_vpadd - 1)

        else:
            # no more room remains; break early.
            break
    # clear text to bottom of screen
    remaining = _endline * (height - line_no)
    output = u''.join((output, remaining))
    return output


def view_article(session, term, url, title):
    """ view an article by target ``url``. """
    # context help
    keyset_help = (u'[r]eturn - (pg)up/down - [{0}%] - [s]hare')

    # display 'fetching ...'
    moveto_lastline = term.move(term.height, 0)
    width = term.width
    if term.kind.startswith('ansi'):
        # bah syncterm
        moveto_lastline = term.move(term.height - 1, 0)
        width -=1
    fetch_txt = 'fetching {0}'.format(url)
    echo(u''.join((moveto_lastline, term.center(fetch_txt[:term.width], width))))

    # perform get request,
    headers = {'User-Agent': USER_AGENT}
    try:
        req = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
    except Exception as err:
        # a wide variety of exceptions may occur; ssl errors, connect timeouts,
        # read errors, BadStatusLine, it goes on and on.
        e_type, _, _ = sys.exc_info()
        echo(u''.join((
            moveto_lastline,
            term.center('{0}: {1}'.format(e_type, err), width))))
        term.inkey()
        return

    if 200 != req.status_code:
        # display 404, 500, or whatever non-200 code returned.
        echo(u''.join((
            moveto_lastline,
            term.center('failed: status_code={0}'.format(req.status_code), width)
        )))
        term.inkey()
        return

    # translate to our session-native encoding,
    req.encoding = session.encoding
    html_text = req.text

    keyset = get_keyset(term)
    _endline = term.clear_eol + u'\r\n'
    bottom = -1
    article_text = u''
    scroll_idx = 0
    last_width, last_height = -1, -1
    do_quit = False
    dirty = True
    width = term.width
    while not do_quit:
        if dirty:
            if last_width != term.width or last_height != term.height:
                # screen size has changed, re-calculate rendered contents
                page_height = term.height - 2
                width = term.width
                if term.kind.startswith('ansi'):
                    # bah syncterm
                    page_height -= 1
                    width -= 1
                article_text = render_article(term, html_text)
                last_width, last_height = term.width, term.height
                bottom = len(article_text) - page_height

            # display titlebar
            echo(term.home)
            echo(getattr(term, COLOR_VIEW)(
                title[:term.width].center(width)))

            # display page at offset
            echo(term.move(1, 0))
            page = slice(scroll_idx, scroll_idx + page_height)
            line_no = 0
            for line_no, line_txt in enumerate(article_text[page]):
                echo(line_txt + _endline)

            # clear (any) remaining lines
            syncterm_sucks = int(not bool(term.kind.startswith('ansi')))
            echo(_endline * (page_height - (line_no + syncterm_sucks)))

            # calculate pct
            _pct = (scroll_idx + page_height) / len(article_text)
            pct = min(int(math.ceil(_pct * 100)), 100)

            # display context help
            moveto_lastline = term.move(term.height, 0)
            if term.kind.startswith('ansi'):
                # bah syncterm
                moveto_lastline = term.move(term.height - 1, 0)
            echo(moveto_lastline)
            echo(getattr(term, COLOR_VIEW)(
                keyset_help.format(pct)[:term.width]
                .center(width)))

            dirty = False

        # block until screen refresh or keyboard input
        event, data = session.read_events(('refresh', 'input'))

        if event == 'refresh':
            dirty = 2
            continue

        elif event == 'input':
            session.buffer_input(data, pushback=True)
            inp = term.inkey(0)
            while inp:
                if inp == u'\x0c':
                    # refresh (^L)
                    dirty = 2
                elif inp.lower() in (u'r', u'q',):
                    do_quit = True
                    break
                elif inp in ('s',):
                    # share url
                    echo(term.move(term.height, 0))
                    echo(term.center(url, width))
                    dirty = True
                    term.inkey()
                else:
                    _idx = do_movement(inp=inp, scroll_idx=scroll_idx,
                                       height=page_height, bottom=bottom,
                                       keyset=keyset)
                    if scroll_idx != _idx:
                        scroll_idx = _idx
                        dirty = 2
                inp = term.inkey(0)


def view_article_summaries(session, term, rss_url, rss_title):
    """ view rss article summary by target ``rss_url``. """
    keyset_help = (u'[q]uit - up/down - [v]iew - [c]omments')

    # fetch rss feed articles
    echo(term.move(term.height // 2, 0))
    echo(term.center('Fetching {0} ...'.format(term.bold(rss_url))).rstrip())
    result = feedparser.parse(rss_url)
    if result.get('status') != 200:
        # display 404, 500, or whatever non-200 code returned.
        moveto_lastline = term.move(term.height, 0)
        echo(moveto_lastline)
        echo(term.center('failed: status={0}'.format(result.get('status'))))
        term.inkey()
        return

    articles = [Article(title=post.title,
                        link=post.link,
                        comments=post.comments,
                        netloc=urlparse.urlparse(post.link).netloc)
                for post in result.entries][:ARTICLE_LIMIT]
    keyset = get_keyset(term)
    bottom = -1
    scroll_idx = 0
    last_width, last_height = -1, -1
    width = term.width
    do_quit = False
    dirty = 2
    while not do_quit:
        if dirty == 2:
            # dirty value of '2' is full screen refresh,
            if last_width != term.width or last_height != term.height:
                # screen size has changed, re-calculate rendered contents
                page_height = term.height - 2
                width = term.width
                if term.kind.startswith('ansi'):
                    # bah syncterm
                    page_height -= 1
                    width -= 1
                last_width, last_height = term.width, term.height

                # we can't actually determine how many articles may be
                # displayed until render_articles_summary is called; we
                # just shortcut and say "the last 3 is the end".
                article_summary = get_article_summaries(term, articles)
                bottom = len(article_summary) - 3

            # display titlebar
            echo(term.home)
            echo(getattr(term, COLOR_MAIN)(
                rss_title[:term.width].center(width)))
            echo(term.move(1, 0))

            # display page summary of articles by scroll index
            echo(render_articles_summary(
                term=term, scroll_idx=scroll_idx,
                height=page_height, articles=article_summary))

        if dirty > 0:
            # dirty value of '1' or more is context-bar refresh
            moveto_lastline = term.move(term.height, 0)
            if term.kind.startswith('ansi'):
                moveto_lastline = term.move(term.height - 1, 0)
            echo(u''.join((
                moveto_lastline,
                getattr(term, COLOR_MAIN)(
                    keyset_help[:term.width].center(width)))
            ))
            dirty = 0

        # block until screen refresh or keyboard input
        event, data = session.read_events(('refresh', 'input'))

        if event == 'refresh':
            dirty = 2
            continue

        elif event == 'input':
            session.buffer_input(data, pushback=True)
            inp = term.inkey(0)
            while inp:
                if inp == u'\x0c':
                    # refresh (^L)
                    dirty = 2
                elif inp.lower() in (u'q',):
                    do_quit = True
                    break
                elif inp in (u'v', 'c'):
                    article = get_article(term, articles)
                    if article is not None:
                        url = {u'v': article.link,
                               u'c': article.comments
                               }[inp]
                        view_article(session=session, term=term,
                                     url=url, title=article.title)
                        dirty = 2
                    else:
                        # refresh only input bar, to occlude only our
                        # prompt for an article (selection invalid)
                        dirty = 1
                    break
                else:
                    _idx = do_movement(inp=inp, scroll_idx=scroll_idx,
                                       height=page_height, bottom=bottom,
                                       keyset=keyset)
                    if scroll_idx != _idx:
                        scroll_idx = _idx
                        dirty = 2
                inp = term.inkey(0)


def main(rss_url=None, rss_title=None):
    """ main entry point. """

    session, term = getsession(), getterminal()
    session.activity = 'reading hackernews'
    rss_url = rss_url or RSS_URL
    rss_title = rss_title or RSS_TITLE

    # move to bottom of screen, reset attribute
    echo(term.move(term.height, 0) + term.normal)

    # create a new, empty screen
    echo(u'\r\n' * (term.height + 1))

    with term.hidden_cursor():
        view_article_summaries(session=session, term=term,
                               rss_url=rss_url, rss_title=rss_title)



"""
SFTP Matrix for x/84.

This script is the default session entry point for all sftp connections.

Nothing especially cute going on here yet.  Ultimately we'd like
a single-directioned communication from x84.sftp.StubSFTPServer
so that we can be notified of the user activity.  We need to make
sure to ensure certain kinds of events are ignored or not handled,
also.

It is *especially* important *never* to echo anything to the
client.  We are still able to push bytes through, though we should
not and allow the sftp subsystem handler to do that for us.

Also, anonymous and new users are not well-handled, here.
"""

# std imports
import logging


def main(anonymous=False, new=False, username=''):
    """ Main procedure. """
    from x84.bbs import (
        getsession,
        getterminal,
        find_user,
        get_user,
        User,
    )

    session, term = getsession(), getterminal()
    session.activity = 'sftp'

    if anonymous:
        user = User(u'anonymous')
    else:
        assert not new, ("new@ user not supported by SFTP.")

        # ProperCase user-specified handle
        handle = find_user(username)
        assert handle is not None, handle

        # fetch user record
        user = get_user(handle)

    # assign session user, just as top.py function login()
    session.user = user

    while True:
        inp = term.inkey()  # should block indefinately
        log = logging.getLogger(__name__)
        log.warn('Got inkey: {0!r}'.format(inp))



#! /usr/bin/env python
""" Distribution file for x/84. """
from __future__ import print_function
import subprocess
import pipes
import errno
import sys
import os

# pylint: disable=E0611,F0401
#         No name 'core' in module 'distutils'
#         Unable to import 'distutils.core'
from distutils.core import setup
from setuptools import Command

HERE = os.path.dirname(__file__)
README = 'README.rst'
DOC_URL = 'http://x84.rtfd.org'


setup(name='x84',
      version='2.0.14',
      description=("Framework for Telnet and SSH BBS or MUD server "
                   "development with example default bbs board"),
      long_description=open(os.path.join(HERE, README)).read(),
      author='Jeff Quast',
      author_email='contact@jeffquast.com',
      url=DOC_URL,
      keywords="telnet ssh terminal server ansi bbs mud curses utf8 cp437",
      license='ISC',
      packages=['x84', 'x84.default', 'x84.default.art', 'x84.bbs',
                'x84.encodings', 'x84.webmodules'],
      package_data={
          '': [README],
          'x84.default': ['*.ans', '*.txt', ],
          'x84.default.art': ['*.asc', '*.ans', '*.txt',
                              'top/*',
                              'weather/*',
                              'bulletins/*/*',
                              ],
      },
      install_requires=[
          'blessed>=1.9.5,<2.0.0',
          'requests==2.5.1',
          'irc==11.0.1',
          'sqlitedict==1.1.0',
          'python-dateutil==2.3',
          'jaraco.timing==1.1',
          'jaraco.util==10.6',
          'more-itertools==2.2',
          'sauce==1.1',
          'six==1.8.0',
          'wsgiref==0.1.2',
          'xmodem==0.3.2',
          'feedparser==5.1.3',
          'html2text==2014.12.29',
      ],
      extras_require={
          'with_crypto': (
              # These cryptogaphy requirements may only be installed:
              # - if a C compiler is available,
              # - if libssl is available,
              # - (sometimes, only) if libffi is available
              #
              # for this reason, they are **optional**, so that x/84 may be
              # installed without a compiler or these external C libraries
              # -- however it is **highly** recommended to always try to
              # use install x84[with_crypto].  It has always been a goal for
              # x/84 to be "pure python" to remain compatible with alternative
              # python interpreter implementations.
              'bcrypt==1.1.0',
              'cffi==1.7.0',
              'cryptography==1.2.3',
              'ecdsa==0.11',
              'enum34==1.0.4',
              'paramiko==1.15.2',
              'pyOpenSSL==0.14',
              'pyasn1==0.1.7',
              'pycparser==2.10',
              'pycrypto==2.6.1',
              'web.py==0.37',
              'cherrypy==3.6.0',
          )
      },
      entry_points={
          'console_scripts': ['x84=x84.engine:main'],
      },
      classifiers=[
          'Environment :: Console :: Curses',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: ISC License (ISCL)',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX :: BSD :: FreeBSD',
          'Operating System :: POSIX :: BSD :: NetBSD',
          'Operating System :: POSIX :: BSD :: OpenBSD',
          'Operating System :: POSIX :: BSD',
          'Operating System :: POSIX :: Linux',
          'Operating System :: POSIX :: SunOS/Solaris',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Programming Language :: Python :: 2 :: Only',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: Artistic Software',
          'Topic :: Communications :: BBS',
          'Topic :: Software Development :: User Interfaces',
          'Topic :: Terminals :: Telnet',
          'Topic :: Terminals',
      ],
      zip_safe=False,
)



#!/usr/bin/env python2.7
""" x84net message poll for x/84. """

# std imports
import logging
import hashlib
import time
import json
import os

# local
from . import cmdline

# 3rd-party
import requests


def get_token(network):
    """ get token for authentication """
    tm_value = int(time.time())
    token = hashlib.sha256('{0}{1}'.format(network['token'], tm_value))
    return '{0}|{1}|{2}'.format(network['board_id'],
                                token.hexdigest(),
                                tm_value)


def prepare_message(msg, network, parent):
    """ turn a Msg object into a dict for transfer """
    from x84.bbs.msgbase import format_origin_line, to_utctime

    return {
        'author': msg.author,
        'subject': msg.subject,
        'recipient': msg.recipient,
        'parent': parent,
        'tags': [tag for tag in msg.tags if tag != network['name']],
        'body': u''.join((msg.body, format_origin_line())),
        'ctime': to_utctime(msg.ctime)
    }


def pull_rest(net, last_msg_id):
    """ pull messages for a given network newer than the 'last' message idx """
    url = '%smessages/%s/%s' % (net['url_base'], net['name'], last_msg_id)

    log = logging.getLogger(__name__)

    try:
        req = requests.get(url,
                           headers={'Auth-X84net': get_token(net)},
                           verify=net['verify'])
    except requests.ConnectionError as err:
        log.warn('[{net[name]}] ConnectionError in pull_rest: {err}'
                 .format(net=net, err=err))
        return False
    except Exception as err:
        log.exception('[{net[name]}] exception in pull_rest: {err}'
                      .format(net=net, err=err))
        return False

    if req.status_code != 200:
        log.error('[{net[name]}] HTTP error, code={req.status_code}'
                  .format(net=net, req=req))
        return False

    try:
        response = json.loads(req.text)
        return response['messages'] if response['response'] else []
    except Exception as err:
        log.exception('[{net[name]}] JSON error: {err}'
                      .format(net=net, err=err))
        return False


def push_rest(net, msg, parent):
    """ push message for a given network and append an origin line """
    msg_data = prepare_message(msg, net, parent)
    url = '{net[url_base]}messages/{net[name]}/'.format(net=net)
    data = {'message': json.dumps(msg_data)}

    log = logging.getLogger(__name__)

    try:
        req = requests.put(url,
                           headers={'Auth-X84net': get_token(net)},
                           data=data,
                           verify=net['verify'])
    except Exception as err:
        log.exception('[{net[name]}] exception in push_rest: {err}'
                      .format(net=net, err=err))
        return False

    if req.status_code not in (200, 201):
        log.error('{net[name]} HTTP error, code={req.status_code}'
                  .format(net=net, req=req))
        return False

    try:
        response = json.loads(req.text)
    except Exception as err:
        log.exception('[{net[name]}] JSON error: {err}'
                      .format(net=net, err=err))
    else:
        if response['response'] and 'id' in response:
            return response['id']
    return False


def get_networks():
    """ Get list configured message networks. """
    from x84.bbs import get_ini

    log = logging.getLogger(__name__)

    # pull list of network-associated tags
    network_list = get_ini(section='msg',
                           key='network_tags',
                           split=True)

    # expected configuration options,
    net_options = ('url_base token board_id'.split())

    networks = list()
    for net_name in network_list:
        net = {'name': net_name}

        section = 'msgnet_{0}'.format(net_name)
        configured = True
        for option in net_options:
            if not get_ini(section=section, key=option):
                log.error('[{net_name}] Missing configuration, '
                          'section=[{section}], option={option}.'
                          .format(net_name=net_name,
                                  section=section,
                                  option=option))
                configured = False
            net[option] = get_ini(section=section, key=option)
        if not configured:
            continue

        # make last_file an absolute path, relative to `datapath`
        net['last_file'] = os.path.join(
            os.path.expanduser(get_ini(section='system', key='datapath')),
            '{net[name]}_last'.format(net=net))

        net['verify'] = True
        ca_path = get_ini(section=section, key='ca_path')
        if ca_path:
            ca_path = os.path.expanduser(ca_path)
            if not os.path.isfile(ca_path):
                log.warn("File not found for Config section [{section}], "
                         "option {key}, value={ca_path}.  default ca_verify "
                         "will be used. ".format(section=section,
                                                 key='ca_path',
                                                 ca_path=ca_path))
            else:
                net['verify'] = ca_path

        networks.append(net)
    return networks


def get_last_msg_id(last_file):
    """ Get the "last message id" by data file ``last_file``. """
    # TODO(jquast): This should have been done internally (and far
    #               more easily!) by a DBProxy database.
    last_msg_id = -1

    log = logging.getLogger(__name__)

    try:
        # May raise IOError (File not Found)
        with open(last_file, 'r') as last_fp:
            last_msg_id = int(last_fp.read().strip())

    except IOError:
        # So, create it; but this too, may raise an
        # OSError (Permission Denied), handled by caller.
        with open(last_file, 'w') as last_fp:
            last_fp.write(str(last_msg_id))

        log.info('last_file created: {0}'.format(last_file))

    return last_msg_id


def poll_network_for_messages(net):
    """ Poll for new messages of network, ``net``. """
    from x84.bbs import Msg, DBProxy
    from x84.bbs.msgbase import to_localtime

    log = logging.getLogger(__name__)

    log.debug(u'[{net[name]}] Polling for new messages.'.format(net=net))

    try:
        last_msg_id = get_last_msg_id(net['last_file'])
    except (OSError, IOError) as err:
        log.error('[{net[name]}] skipping network: {err}'
                  .format(net=net, err=err))
        return

    msgs = pull_rest(net=net, last_msg_id=last_msg_id)

    if msgs:
        log.info('[{net[name]}] Retrieved {num} messages.'
                 .format(net=net, num=len(msgs)))
    else:
        log.debug('[{net[name]}] No messages.'.format(net=net))
        return

    transdb = DBProxy('{0}trans'.format(net['name']), use_session=False)
    transkeys = transdb.keys()
    msgs = sorted(msgs, cmp=lambda x, y: cmp(int(x['id']), int(y['id'])))

    # store messages locally, saving their translated IDs to the transdb
    for msg in msgs:
        store_msg = Msg()
        store_msg.recipient = msg['recipient']
        store_msg.author = msg['author']
        store_msg.subject = msg['subject']
        store_msg.body = msg['body']
        store_msg.tags = set(msg['tags'])
        store_msg.tags.add(u''.join((net['name'])))

        if msg['recipient'] is None and u'public' not in msg['tags']:
            log.warn("[{net[name]}] No recipient (msg_id={msg[id]}), "
                     "adding 'public' tag".format(net=net, msg=msg))
            store_msg.tags.add(u'public')

        if (msg['parent'] is not None and
                str(msg['parent']) not in transkeys):
            log.warn('[{net[name]}] No such parent message ({msg[parent]}, '
                     'msg_id={msg[id]}), removing reference.'
                     .format(net=net, msg=msg))
        elif msg['parent'] is not None:
            store_msg.parent = int(transdb[msg['parent']])

        if msg['id'] in transkeys:
            log.warn('[{net[name]}] dupe (msg_id={msg[id]}) discarded.'
                     .format(net=net, msg=msg))
        else:
            # do not save this message to network, we already received
            # it from the network, set send_net=False
            store_msg.save(send_net=False, ctime=to_localtime(msg['ctime']))
            with transdb:
                transdb[msg['id']] = store_msg.idx
            transkeys.append(msg['id'])
            log.info('[{net[name]}] Processed (msg_id={msg[id]}) => {new_id}'
                     .format(net=net, msg=msg, new_id=store_msg.idx))

        if 'last' not in net.keys() or int(net['last']) < int(msg['id']):
            net['last'] = msg['id']

    if 'last' in net.keys():
        with open(net['last_file'], 'w') as last_fp:
            last_fp.write(str(net['last']))

    return


def publish_network_messages(net):
    """ Push messages to network, ``net``. """
    from x84.bbs import DBProxy
    from x84.bbs.msgbase import format_origin_line, MSGDB

    log = logging.getLogger(__name__)

    log.debug(u'[{net[name]}] publishing new messages.'.format(net=net))

    queuedb = DBProxy('{0}queues'.format(net['name']), use_session=False)
    transdb = DBProxy('{0}trans'.format(net['name']), use_session=False)
    msgdb = DBProxy(MSGDB, use_session=False)

    # publish each message
    for msg_id in sorted(queuedb.keys(),
                         cmp=lambda x, y: cmp(int(x), int(y))):
        if msg_id not in msgdb:
            log.warn('[{net[name]}] No such message (msg_id={msg_id})'
                     .format(net=net, msg_id=msg_id))
            del queuedb[msg_id]
            continue

        msg = msgdb[msg_id]

        trans_parent = None
        if msg.parent is not None:
            matches = [key for key, data in transdb.items()
                       if int(data) == msg.parent]

            if len(matches) > 0:
                trans_parent = matches[0]
            else:
                log.warn('[{net[name]}] Parent ID {msg.parent} '
                         'not in translation-DB (msg_id={msg_id})'
                         .format(net=net, msg=msg, msg_id=msg_id))

        trans_id = push_rest(net=net, msg=msg, parent=trans_parent)
        if trans_id is False:
            log.error('[{net[name]}] Message not posted (msg_id={msg_id})'
                      .format(net=net, msg_id=msg_id))
            continue

        if trans_id in transdb.keys():
            log.error('[{net[name]}] trans_id={trans_id} conflicts with '
                      '(msg_id={msg_id})'
                      .format(net=net, trans_id=trans_id, msg_id=msg_id))
            with queuedb:
                del queuedb[msg_id]
            continue

        # transform, and possibly duplicate(?) message ..
        with transdb, msgdb, queuedb:
            transdb[trans_id] = msg_id
            msg.body = u''.join((msg.body, format_origin_line()))
            msgdb[msg_id] = msg
            del queuedb[msg_id]
        log.info('[{net[name]}] Published (msg_id={msg_id}) => {trans_id}'
                 .format(net=net, msg_id=msg_id, trans_id=trans_id))


def poller(poll_interval):
    """ Blocking function periodically polls configured message networks. """
    log = logging.getLogger(__name__)

    # get all networks
    networks = get_networks()

    if networks:
        while True:
            do_poll(networks)
            time.sleep(poll_interval)
    else:
        log.error(u'No networks configured for poll/publish.')


def main(background_daemon=True):
    """
    Entry point to configure and begin network message polling.

    Called by x84/engine.py, function main() as unmanaged thread.

    :param bool background_daemon: When True (default), this function returns
                and web modules are served in an unmanaged, background (daemon)
                thread.  Otherwise, function call to ``main()`` is blocking.
    :rtype: None
    """
    from threading import Thread
    from x84.bbs.ini import get_ini

    log = logging.getLogger(__name__)

    poll_interval = get_ini(section='msg',
                            key='poll_interval',
                            getter='getint'
                            ) or 1984

    if background_daemon:
        t = Thread(target=poller, args=(poll_interval,))
        t.daemon = True
        log.info('msgpoll at {0}s intervals.'.format(poll_interval))
        t.start()
    else:
        poller(poll_interval)


def do_poll(networks):
    """
    Message polling process.

    Function is called periodically by :func:`poller`.
    """
    # pull-from all networks
    map(poll_network_for_messages, networks)

    # publish-to all networks
    map(publish_network_messages, networks)

if __name__ == '__main__':
    # load only message polling module when executing this script directly.
    #
    # as we are running outside of the 'engine' context, it is necessary
    # for us to initialize the .ini configuration scheme so that the list
    # of web modules and ssl options may be gathered.
    import x84.bbs.ini
    x84.bbs.ini.init(*cmdline.parse_args())

    # do not execute message polling as a background thread.
    main(background_daemon=False)



"""
Atari ST codec for x/84.

Generated from:
ftp://ftp.unicode.org/Public/MAPPINGS/VENDORS/MISC/ATARIST.TXT
"""

import codecs

# Codec APIs


class Codec(codecs.Codec):

    def encode(self, char, errors='strict'):
        return codecs.charmap_encode(char, errors, ENCODING_TABLE)

    def decode(self, char, errors='strict'):
        return codecs.charmap_decode(char, errors, DECODING_TABLE)


class IncrementalEncoder(codecs.IncrementalEncoder):

    def encode(self, char, final=False):
        return codecs.charmap_encode(char, self.errors, ENCODING_TABLE)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):

    def decode(self, char, final=False):
        return codecs.charmap_decode(char, self.errors, DECODING_TABLE)[0]


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass


# encodings module API

def getaliases():
    return ('atari',)


def getregentry():
    return codecs.CodecInfo(
        name='atarist',
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


# Decoding Table

DECODING_TABLE = (
    u'\x00'  # 0x00 -> NULL
    u'\x01'  # 0x01 -> START OF HEADING
    u'\x02'  # 0x02 -> START OF TEXT
    u'\x03'  # 0x03 -> END OF TEXT
    u'\x04'  # 0x04 -> END OF TRANSMISSION
    u'\x05'  # 0x05 -> ENQUIRY
    u'\x06'  # 0x06 -> ACKNOWLEDGE
    u'\x07'  # 0x07 -> BELL
    u'\x08'  # 0x08 -> BACKSPACE
    u'\t'  # 0x09 -> HORIZONTAL TABULATION
    u'\n'  # 0x0A -> LINE FEED
    u'\x0b'  # 0x0B -> VERTICAL TABULATION
    u'\x0c'  # 0x0C -> FORM FEED
    u'\r'  # 0x0D -> CARRIAGE RETURN
    u'\x0e'  # 0x0E -> SHIFT OUT
    u'\x0f'  # 0x0F -> SHIFT IN
    u'\x10'  # 0x10 -> DATA LINK ESCAPE
    u'\x11'  # 0x11 -> DEVICE CONTROL ONE
    u'\x12'  # 0x12 -> DEVICE CONTROL TWO
    u'\x13'  # 0x13 -> DEVICE CONTROL THREE
    u'\x14'  # 0x14 -> DEVICE CONTROL FOUR
    u'\x15'  # 0x15 -> NEGATIVE ACKNOWLEDGE
    u'\x16'  # 0x16 -> SYNCHRONOUS IDLE
    u'\x17'  # 0x17 -> END OF TRANSMISSION BLOCK
    u'\x18'  # 0x18 -> CANCEL
    u'\x19'  # 0x19 -> END OF MEDIUM
    u'\x1a'  # 0x1A -> SUBSTITUTE
    u'\x1b'  # 0x1B -> ESCAPE
    u'\x1c'  # 0x1C -> FILE SEPARATOR
    u'\x1d'  # 0x1D -> GROUP SEPARATOR
    u'\x1e'  # 0x1E -> RECORD SEPARATOR
    u'\x1f'  # 0x1F -> UNIT SEPARATOR
    u' '  # 0x20 -> SPACE
    u'!'  # 0x21 -> EXCLAMATION MARK
    u'"'  # 0x22 -> QUOTATION MARK
    u'#'  # 0x23 -> NUMBER SIGN
    u'$'  # 0x24 -> DOLLAR SIGN
    u'%'  # 0x25 -> PERCENT SIGN
    u'&'  # 0x26 -> AMPERSAND
    u"'"  # 0x27 -> APOSTROPHE
    u'('  # 0x28 -> LEFT PARENTHESIS
    u')'  # 0x29 -> RIGHT PARENTHESIS
    u'*'  # 0x2A -> ASTERISK
    u'+'  # 0x2B -> PLUS SIGN
    u','  # 0x2C -> COMMA
    u'-'  # 0x2D -> HYPHEN-MINUS
    u'.'  # 0x2E -> FULL STOP
    u'/'  # 0x2F -> SOLIDUS
    u'0'  # 0x30 -> DIGIT ZERO
    u'1'  # 0x31 -> DIGIT ONE
    u'2'  # 0x32 -> DIGIT TWO
    u'3'  # 0x33 -> DIGIT THREE
    u'4'  # 0x34 -> DIGIT FOUR
    u'5'  # 0x35 -> DIGIT FIVE
    u'6'  # 0x36 -> DIGIT SIX
    u'7'  # 0x37 -> DIGIT SEVEN
    u'8'  # 0x38 -> DIGIT EIGHT
    u'9'  # 0x39 -> DIGIT NINE
    u':'  # 0x3A -> COLON
    u';'  # 0x3B -> SEMICOLON
    u'<'  # 0x3C -> LESS-THAN SIGN
    u'='  # 0x3D -> EQUALS SIGN
    u'>'  # 0x3E -> GREATER-THAN SIGN
    u'?'  # 0x3F -> QUESTION MARK
    u'@'  # 0x40 -> COMMERCIAL AT
    u'A'  # 0x41 -> LATIN CAPITAL LETTER A
    u'B'  # 0x42 -> LATIN CAPITAL LETTER B
    u'C'  # 0x43 -> LATIN CAPITAL LETTER C
    u'D'  # 0x44 -> LATIN CAPITAL LETTER D
    u'E'  # 0x45 -> LATIN CAPITAL LETTER E
    u'F'  # 0x46 -> LATIN CAPITAL LETTER F
    u'G'  # 0x47 -> LATIN CAPITAL LETTER G
    u'H'  # 0x48 -> LATIN CAPITAL LETTER H
    u'I'  # 0x49 -> LATIN CAPITAL LETTER I
    u'J'  # 0x4A -> LATIN CAPITAL LETTER J
    u'K'  # 0x4B -> LATIN CAPITAL LETTER K
    u'L'  # 0x4C -> LATIN CAPITAL LETTER L
    u'M'  # 0x4D -> LATIN CAPITAL LETTER M
    u'N'  # 0x4E -> LATIN CAPITAL LETTER N
    u'O'  # 0x4F -> LATIN CAPITAL LETTER O
    u'P'  # 0x50 -> LATIN CAPITAL LETTER P
    u'Q'  # 0x51 -> LATIN CAPITAL LETTER Q
    u'R'  # 0x52 -> LATIN CAPITAL LETTER R
    u'S'  # 0x53 -> LATIN CAPITAL LETTER S
    u'T'  # 0x54 -> LATIN CAPITAL LETTER T
    u'U'  # 0x55 -> LATIN CAPITAL LETTER U
    u'V'  # 0x56 -> LATIN CAPITAL LETTER V
    u'W'  # 0x57 -> LATIN CAPITAL LETTER W
    u'X'  # 0x58 -> LATIN CAPITAL LETTER X
    u'Y'  # 0x59 -> LATIN CAPITAL LETTER Y
    u'Z'  # 0x5A -> LATIN CAPITAL LETTER Z
    u'['  # 0x5B -> LEFT SQUARE BRACKET
    u'\\'  # 0x5C -> REVERSE SOLIDUS
    u']'  # 0x5D -> RIGHT SQUARE BRACKET
    u'^'  # 0x5E -> CIRCUMFLEX ACCENT
    u'_'  # 0x5F -> LOW LINE
    u'`'  # 0x60 -> GRAVE ACCENT
    u'a'  # 0x61 -> LATIN SMALL LETTER A
    u'b'  # 0x62 -> LATIN SMALL LETTER B
    u'c'  # 0x63 -> LATIN SMALL LETTER C
    u'd'  # 0x64 -> LATIN SMALL LETTER D
    u'e'  # 0x65 -> LATIN SMALL LETTER E
    u'f'  # 0x66 -> LATIN SMALL LETTER F
    u'g'  # 0x67 -> LATIN SMALL LETTER G
    u'h'  # 0x68 -> LATIN SMALL LETTER H
    u'i'  # 0x69 -> LATIN SMALL LETTER I
    u'j'  # 0x6A -> LATIN SMALL LETTER J
    u'k'  # 0x6B -> LATIN SMALL LETTER K
    u'l'  # 0x6C -> LATIN SMALL LETTER L
    u'm'  # 0x6D -> LATIN SMALL LETTER M
    u'n'  # 0x6E -> LATIN SMALL LETTER N
    u'o'  # 0x6F -> LATIN SMALL LETTER O
    u'p'  # 0x70 -> LATIN SMALL LETTER P
    u'q'  # 0x71 -> LATIN SMALL LETTER Q
    u'r'  # 0x72 -> LATIN SMALL LETTER R
    u's'  # 0x73 -> LATIN SMALL LETTER S
    u't'  # 0x74 -> LATIN SMALL LETTER T
    u'u'  # 0x75 -> LATIN SMALL LETTER U
    u'v'  # 0x76 -> LATIN SMALL LETTER V
    u'w'  # 0x77 -> LATIN SMALL LETTER W
    u'x'  # 0x78 -> LATIN SMALL LETTER X
    u'y'  # 0x79 -> LATIN SMALL LETTER Y
    u'z'  # 0x7A -> LATIN SMALL LETTER Z
    u'{'  # 0x7B -> LEFT CURLY BRACKET
    u'|'  # 0x7C -> VERTICAL LINE
    u'}'  # 0x7D -> RIGHT CURLY BRACKET
    u'~'  # 0x7E -> TILDE
    u'\x7f'  # 0x7F -> DELETE
    u'\xc7'  # 0x80 -> LATIN CAPITAL LETTER C WITH CEDILLA
    u'\xfc'  # 0x81 -> LATIN SMALL LETTER U WITH DIAERESIS
    u'\xe9'  # 0x82 -> LATIN SMALL LETTER E WITH ACUTE
    u'\xe2'  # 0x83 -> LATIN SMALL LETTER A WITH CIRCUMFLEX
    u'\xe4'  # 0x84 -> LATIN SMALL LETTER A WITH DIAERESIS
    u'\xe0'  # 0x85 -> LATIN SMALL LETTER A WITH GRAVE
    u'\xe5'  # 0x86 -> LATIN SMALL LETTER A WITH RING ABOVE
    u'\xe7'  # 0x87 -> LATIN SMALL LETTER C WITH CEDILLA
    u'\xea'  # 0x88 -> LATIN SMALL LETTER E WITH CIRCUMFLEX
    u'\xeb'  # 0x89 -> LATIN SMALL LETTER E WITH DIAERESIS
    u'\xe8'  # 0x8A -> LATIN SMALL LETTER E WITH GRAVE
    u'\xef'  # 0x8B -> LATIN SMALL LETTER I WITH DIAERESIS
    u'\xee'  # 0x8C -> LATIN SMALL LETTER I WITH CIRCUMFLEX
    u'\xec'  # 0x8D -> LATIN SMALL LETTER I WITH GRAVE
    u'\xc4'  # 0x8E -> LATIN CAPITAL LETTER A WITH DIAERESIS
    u'\xc5'  # 0x8F -> LATIN CAPITAL LETTER A WITH RING ABOVE
    u'\xc9'  # 0x90 -> LATIN CAPITAL LETTER E WITH ACUTE
    u'\xe6'  # 0x91 -> LATIN SMALL LETTER AE
    u'\xc6'  # 0x92 -> LATIN CAPITAL LETTER AE
    u'\xf4'  # 0x93 -> LATIN SMALL LETTER O WITH CIRCUMFLEX
    u'\xf6'  # 0x94 -> LATIN SMALL LETTER O WITH DIAERESIS
    u'\xf2'  # 0x95 -> LATIN SMALL LETTER O WITH GRAVE
    u'\xfb'  # 0x96 -> LATIN SMALL LETTER U WITH CIRCUMFLEX
    u'\xf9'  # 0x97 -> LATIN SMALL LETTER U WITH GRAVE
    u'\xff'  # 0x98 -> LATIN SMALL LETTER Y WITH DIAERESIS
    u'\xd6'  # 0x99 -> LATIN CAPITAL LETTER O WITH DIAERESIS
    u'\xdc'  # 0x9A -> LATIN CAPITAL LETTER U WITH DIAERESIS
    u'\xa2'  # 0x9B -> CENT SIGN
    u'\xa3'  # 0x9C -> POUND SIGN
    u'\xa5'  # 0x9D -> YEN SIGN
    u'\xdf'  # 0x9E -> LATIN SMALL LETTER SHARP S
    u'\u0192'  # 0x9F -> LATIN SMALL LETTER F WITH HOOK
    u'\xe1'  # 0xA0 -> LATIN SMALL LETTER A WITH ACUTE
    u'\xed'  # 0xA1 -> LATIN SMALL LETTER I WITH ACUTE
    u'\xf3'  # 0xA2 -> LATIN SMALL LETTER O WITH ACUTE
    u'\xfa'  # 0xA3 -> LATIN SMALL LETTER U WITH ACUTE
    u'\xf1'  # 0xA4 -> LATIN SMALL LETTER N WITH TILDE
    u'\xd1'  # 0xA5 -> LATIN CAPITAL LETTER N WITH TILDE
    u'\xaa'  # 0xA6 -> FEMININE ORDINAL INDICATOR
    u'\xba'  # 0xA7 -> MASCULINE ORDINAL INDICATOR
    u'\xbf'  # 0xA8 -> INVERTED QUESTION MARK
    u'\u2310'  # 0xA9 -> REVERSED NOT SIGN
    u'\xac'  # 0xAA -> NOT SIGN
    u'\xbd'  # 0xAB -> VULGAR FRACTION ONE HALF
    u'\xbc'  # 0xAC -> VULGAR FRACTION ONE QUARTER
    u'\xa1'  # 0xAD -> INVERTED EXCLAMATION MARK
    u'\xab'  # 0xAE -> LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    u'\xbb'  # 0xAF -> RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    u'\xe3'  # 0xB0 -> LATIN SMALL LETTER A WITH TILDE
    u'\xf5'  # 0xB1 -> LATIN SMALL LETTER O WITH TILDE
    u'\xd8'  # 0xB2 -> LATIN CAPITAL LETTER O WITH STROKE
    u'\xf8'  # 0xB3 -> LATIN SMALL LETTER O WITH STROKE
    u'\u0153'  # 0xB4 -> LATIN SMALL LIGATURE OE
    u'\u0152'  # 0xB5 -> LATIN CAPITAL LIGATURE OE
    u'\xc0'  # 0xB6 -> LATIN CAPITAL LETTER A WITH GRAVE
    u'\xc3'  # 0xB7 -> LATIN CAPITAL LETTER A WITH TILDE
    u'\xd5'  # 0xB8 -> LATIN CAPITAL LETTER O WITH TILDE
    u'\xa8'  # 0xB9 -> DIAERESIS
    u'\xb4'  # 0xBA -> ACUTE ACCENT
    u'\u2020'  # 0xBB -> DAGGER
    u'\xb6'  # 0xBC -> PILCROW SIGN
    u'\xa9'  # 0xBD -> COPYRIGHT SIGN
    u'\xae'  # 0xBE -> REGISTERED SIGN
    u'\u2122'  # 0xBF -> TRADE MARK SIGN
    u'\u0133'  # 0xC0 -> LATIN SMALL LIGATURE IJ
    u'\u0132'  # 0xC1 -> LATIN CAPITAL LIGATURE IJ
    u'\u05d0'  # 0xC2 -> HEBREW LETTER ALEF
    u'\u05d1'  # 0xC3 -> HEBREW LETTER BET
    u'\u05d2'  # 0xC4 -> HEBREW LETTER GIMEL
    u'\u05d3'  # 0xC5 -> HEBREW LETTER DALET
    u'\u05d4'  # 0xC6 -> HEBREW LETTER HE
    u'\u05d5'  # 0xC7 -> HEBREW LETTER VAV
    u'\u05d6'  # 0xC8 -> HEBREW LETTER ZAYIN
    u'\u05d7'  # 0xC9 -> HEBREW LETTER HET
    u'\u05d8'  # 0xCA -> HEBREW LETTER TET
    u'\u05d9'  # 0xCB -> HEBREW LETTER YOD
    u'\u05db'  # 0xCC -> HEBREW LETTER KAF
    u'\u05dc'  # 0xCD -> HEBREW LETTER LAMED
    u'\u05de'  # 0xCE -> HEBREW LETTER MEM
    u'\u05e0'  # 0xCF -> HEBREW LETTER NUN
    u'\u05e1'  # 0xD0 -> HEBREW LETTER SAMEKH
    u'\u05e2'  # 0xD1 -> HEBREW LETTER AYIN
    u'\u05e4'  # 0xD2 -> HEBREW LETTER PE
    u'\u05e6'  # 0xD3 -> HEBREW LETTER TSADI
    u'\u05e7'  # 0xD4 -> HEBREW LETTER QOF
    u'\u05e8'  # 0xD5 -> HEBREW LETTER RESH
    u'\u05e9'  # 0xD6 -> HEBREW LETTER SHIN
    u'\u05ea'  # 0xD7 -> HEBREW LETTER TAV
    u'\u05df'  # 0xD8 -> HEBREW LETTER FINAL NUN
    u'\u05da'  # 0xD9 -> HEBREW LETTER FINAL KAF
    u'\u05dd'  # 0xDA -> HEBREW LETTER FINAL MEM
    u'\u05e3'  # 0xDB -> HEBREW LETTER FINAL PE
    u'\u05e5'  # 0xDC -> HEBREW LETTER FINAL TSADI
    u'\xa7'  # 0xDD -> SECTION SIGN
    u'\u2227'  # 0xDE -> LOGICAL AND
    u'\u221e'  # 0xDF -> INFINITY
    u'\u03b1'  # 0xE0 -> GREEK SMALL LETTER ALPHA
    u'\u03b2'  # 0xE1 -> GREEK SMALL LETTER BETA
    u'\u0393'  # 0xE2 -> GREEK CAPITAL LETTER GAMMA
    u'\u03c0'  # 0xE3 -> GREEK SMALL LETTER PI
    u'\u03a3'  # 0xE4 -> GREEK CAPITAL LETTER SIGMA
    u'\u03c3'  # 0xE5 -> GREEK SMALL LETTER SIGMA
    u'\xb5'  # 0xE6 -> MICRO SIGN
    u'\u03c4'  # 0xE7 -> GREEK SMALL LETTER TAU
    u'\u03a6'  # 0xE8 -> GREEK CAPITAL LETTER PHI
    u'\u0398'  # 0xE9 -> GREEK CAPITAL LETTER THETA
    u'\u03a9'  # 0xEA -> GREEK CAPITAL LETTER OMEGA
    u'\u03b4'  # 0xEB -> GREEK SMALL LETTER DELTA
    u'\u222e'  # 0xEC -> CONTOUR INTEGRAL
    u'\u03c6'  # 0xED -> GREEK SMALL LETTER PHI
    u'\u2208'  # 0xEE -> ELEMENT OF SIGN
    u'\u2229'  # 0xEF -> INTERSECTION
    u'\u2261'  # 0xF0 -> IDENTICAL TO
    u'\xb1'  # 0xF1 -> PLUS-MINUS SIGN
    u'\u2265'  # 0xF2 -> GREATER-THAN OR EQUAL TO
    u'\u2264'  # 0xF3 -> LESS-THAN OR EQUAL TO
    u'\u2320'  # 0xF4 -> TOP HALF INTEGRAL
    u'\u2321'  # 0xF5 -> BOTTOM HALF INTEGRAL
    u'\xf7'  # 0xF6 -> DIVISION SIGN
    u'\u2248'  # 0xF7 -> ALMOST EQUAL TO
    u'\xb0'  # 0xF8 -> DEGREE SIGN
    u'\u2219'  # 0xF9 -> BULLET OPERATOR
    u'\xb7'  # 0xFA -> MIDDLE DOT
    u'\u221a'  # 0xFB -> SQUARE ROOT
    u'\u207f'  # 0xFC -> SUPERSCRIPT LATIN SMALL LETTER N
    u'\xb2'  # 0xFD -> SUPERSCRIPT TWO
    u'\xb3'  # 0xFE -> SUPERSCRIPT THREE
    u'\xaf'  # 0xFF -> MACRON
)

# Encoding table
ENCODING_TABLE = codecs.charmap_build(DECODING_TABLE)



""" Configuration package x/84. """
# std imports
from __future__ import print_function
import logging.config
import ConfigParser
import warnings
import inspect
import getpass
import socket
import os

#: Singleton representing configuration after load
CFG = None

# pylint: disable=R0915,R0912,W0603
#         Too many statements
#         Too many branches
#         Using the global statement


def init(lookup_bbs, lookup_log):
    """
    Initialize global 'CFG' variable, a singleton to contain bbs settings.

    Each variable (``lookup_bbs``, ``lookup_log``) is tuple lookup path of
    in-order preferences for .ini files.  If none are found, defaults are
    initialized, and the last item of each tuple is created.
    """
    log = logging.getLogger(__name__)

    def write_cfg(cfg, filepath):
        """ Write Config to filepath. """
        if not os.path.exists(os.path.dirname(os.path.expanduser(filepath))):
            dir_name = os.path.dirname(os.path.expanduser(filepath))
            print('Creating folder {0}'.format(dir_name))
            os.mkdir(dir_name)
        print('Saving {0}'.format(filepath))
        cfg.write(open(os.path.expanduser(filepath), 'w'))

    # exploit last argument, presumed to be within a folder
    # writable by our process, and where the ini is wanted
    # -- engine.py specifys a default of: ~/.x84/somefile.ini
    loaded = False
    cfg_logfile = lookup_log[-1]
    for cfg_logfile in lookup_log:
        cfg_logfile = os.path.expanduser(cfg_logfile)
        # load-only defaults,
        if os.path.exists(cfg_logfile):
            print('loading {0}'.format((cfg_logfile)))
            logging.config.fileConfig(cfg_logfile)
            loaded = True
            break
    if not loaded:
        cfg_log = init_log_ini()
        dir_name = os.path.dirname(cfg_logfile)
        if not os.path.isdir(dir_name):
            try:
                os.makedirs(dir_name)
            except OSError as err:
                log.warn(err)
        try:
            write_cfg(cfg_log, cfg_logfile)
            log.info('Saved %s', cfg_logfile)
        except IOError as err:
            log.error(err)
        logging.config.fileConfig(cfg_logfile)

    loaded = False
    cfg_bbs = ConfigParser.SafeConfigParser()
    cfg_bbsfile = lookup_bbs[-1]
    for cfg_bbsfile in lookup_bbs:
        cfg_bbsfile = os.path.expanduser(cfg_bbsfile)
        # load defaults,
        if os.path.exists(cfg_bbsfile):
            cfg_bbs.read(cfg_bbsfile)
            log.info('loaded %s', cfg_bbsfile)
            loaded = True
            break
    if not loaded:
        cfg_bbs = init_bbs_ini()
        dir_name = os.path.dirname(cfg_bbsfile)
        if not os.path.isdir(dir_name):
            try:
                os.makedirs(dir_name)
            except OSError as err:
                log.warn(err)
        try:
            write_cfg(cfg_bbs, cfg_bbsfile)
            log.info('Saved %s', cfg_bbsfile)
        except IOError as err:
            log.error(err)

    global CFG
    CFG = cfg_bbs


def init_bbs_ini():
    """ Returns ConfigParser instance of bbs system defaults. """
    # ### How this should have been written ...
    # ###
    # ### each module should provide a declaration, probably based on a
    # ### collections.namedtuple that simply states the various fields,
    # ### a help description, and its default values.  Then, this program
    # ### simply "walks" the module path, importing everybody and finding
    # ### this declaration to build up the "default" configuration file.
    # ###
    # ### at least, in this way, the module defines its configuration scheme
    # ### where it is used.
    # wouldn't it be nice if we could use comments in the file .. ?
    # in such cases, it might be better to use jinja2 or something
    cfg_bbs = ConfigParser.SafeConfigParser()

    cfg_bbs.add_section('system')
    cfg_bbs.set('system', 'bbsname', 'x/84')
    cfg_bbs.set('system', 'sysop', '')
    cfg_bbs.set('system', 'software', 'x/84')
    # use module-level 'default' folder
    cfg_bbs.set('system', 'scriptpath', os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.path.pardir, 'default')))
    cfg_bbs.set('system', 'datapath', os.path.expanduser(os.path.join(
        os.path.join('~', '.x84', 'data'))))
    cfg_bbs.set('system', 'timeout', '1984')

    try:
        # pylint: disable=W0612
        #         Unused variable 'bcrypt'
        import bcrypt  # NOQA
    except ImportError:
        cfg_bbs.set('system', 'password_digest', 'internal')
    else:
        cfg_bbs.set('system', 'password_digest', 'bcrypt')
    cfg_bbs.set('system', 'mail_addr',
                '%s@%s' % (getpass.getuser(), socket.gethostname()))
    cfg_bbs.set('system', 'mail_smtphost', 'localhost')

    # one *Could* change 'ansi' termcaps to 'ansi-bbs', for SynchTerm,
    # but how do we identify that 'ansi-bbs' TERM is available on this
    # system? hmm .. lets offer the reverse, anything beginning with
    # 'ansi' can changed to any other value; so we could be
    # unidirectional: a value of 'ansi' will translate ansi-bbs -> ansi,
    # and a value of 'ansi-bbs' will translate ansi -> ansi-bbs.
    cfg_bbs.set('system', 'termcap-ansi', 'ansi')
    # change 'unknown' termcaps to 'ansi': for dumb terminals
    cfg_bbs.set('system', 'termcap-unknown', 'ansi')
    # could be information leak to sensitive sysops
    cfg_bbs.set('system', 'show_traceback', 'yes')
    # store passwords in uppercase, facebook and mystic bbs does this ..
    cfg_bbs.set('system', 'pass_ucase', 'no')
    # default encoding for the showart function on UTF-8 capable terminals
    cfg_bbs.set('system', 'art_utf8_codec', 'cp437')

    cfg_bbs.add_section('telnet')
    cfg_bbs.set('telnet', 'enabled', 'yes')
    cfg_bbs.set('telnet', 'addr', '127.0.0.1')
    cfg_bbs.set('telnet', 'port', '6023')

    cfg_bbs.add_section('ssh')
    try:
        # pylint: disable=W0612
        #         Unused variable 'x84'
        import x84.ssh  # noqa
        cfg_bbs.set('ssh', 'enabled', 'yes')
    except ImportError:
        cfg_bbs.set('ssh', 'enabled', 'no')
    cfg_bbs.set('ssh', 'addr', '127.0.0.1')
    cfg_bbs.set('ssh', 'port', '6022')
    cfg_bbs.set('ssh', 'hostkey', os.path.expanduser(
        os.path.join('~', '.x84', 'ssh_host_rsa_key')))
    cfg_bbs.set('ssh', 'hostkeybits', '2048')

    cfg_bbs.add_section('sftp')
    cfg_bbs.set('sftp', 'enabled', 'no')
    cfg_bbs.set('sftp', 'root', os.path.expanduser(
        os.path.join('~', 'x84-sftp_root')))
    try:
        os.makedirs(
            os.path.join(cfg_bbs.get('sftp', 'root'), "__uploads__"))
    except OSError:
        pass
    cfg_bbs.set('sftp', 'uploads_filemode', '644')

    # rlogin only works on port 513
    cfg_bbs.add_section('rlogin')
    cfg_bbs.set('rlogin', 'enabled', 'no')
    cfg_bbs.set('rlogin', 'addr', '127.0.0.1')
    cfg_bbs.set('rlogin', 'port', '513')

    # web
    cfg_bbs.add_section('web')
    cfg_bbs.set('web', 'enabled', 'no')
    cfg_bbs.set('web', 'port', '443')
    cfg_bbs.set('web', 'cert', os.path.expanduser(
        os.path.join('~', '.x84', 'ssl.cer')))
    cfg_bbs.set('web', 'key', os.path.expanduser(
        os.path.join('~', '.x84', 'ssl.key')))
    cfg_bbs.set('web', 'chain', os.path.expanduser(
        os.path.join('~', '.x84', 'ca.cer')))
    cfg_bbs.set('web', 'modules', 'msgserve')

    # default path if cmd argument is not absolute,
    cfg_bbs.add_section('door')
    cfg_bbs.set('door', 'path', '/usr/local/bin:/usr/games')

    cfg_bbs.add_section('matrix')
    cfg_bbs.set('matrix', 'newcmds', 'new, apply')
    cfg_bbs.set('matrix', 'byecmds', 'exit, logoff, bye, quit')
    cfg_bbs.set('matrix', 'anoncmds', 'anonymous')
    cfg_bbs.set('matrix', 'script', 'matrix')
    cfg_bbs.set('matrix', 'script_telnet', 'matrix')
    cfg_bbs.set('matrix', 'script_ssh', 'matrix_ssh')
    cfg_bbs.set('matrix', 'script_sftp', 'matrix_sftp')
    cfg_bbs.set('matrix', 'topscript', 'top')
    cfg_bbs.set('matrix', 'enable_anonymous', 'no')
    cfg_bbs.set('matrix', 'enable_pwreset', 'yes')

    cfg_bbs.add_section('session')
    cfg_bbs.set('session', 'tap_input', 'no')
    cfg_bbs.set('session', 'tap_output', 'no')
    cfg_bbs.set('session', 'tap_events', 'no')
    cfg_bbs.set('session', 'tap_db', 'no')
    cfg_bbs.set('session', 'default_encoding', 'utf8')

    cfg_bbs.add_section('irc')
    cfg_bbs.set('irc', 'server', 'efnet.portlane.se')
    cfg_bbs.set('irc', 'port', '6667')
    cfg_bbs.set('irc', 'channel', '#1984')
    cfg_bbs.set('irc', 'enable_privnotice', 'yes')
    cfg_bbs.set('irc', 'maxnick', '9')
    cfg_bbs.set('irc', 'ssl', 'no')

    cfg_bbs.add_section('shroo-ms')
    cfg_bbs.set('shroo-ms', 'enabled', 'no')
    cfg_bbs.set('shroo-ms', 'idkey', '')
    cfg_bbs.set('shroo-ms', 'restkey', '')

    # new user account script
    cfg_bbs.add_section('nua')
    cfg_bbs.set('nua', 'script', 'nua')
    cfg_bbs.set('nua', 'min_user', '3')
    cfg_bbs.set('nua', 'min_pass', '4')
    cfg_bbs.set('nua', 'max_user', '11')
    cfg_bbs.set('nua', 'max_pass', '16')
    cfg_bbs.set('nua', 'max_email', '30')
    cfg_bbs.set('nua', 'max_location', '24')
    cfg_bbs.set('nua', 'allow_apply', 'yes')
    invalid_handles = u', '.join((
        cfg_bbs.get('matrix', 'byecmds'),
        cfg_bbs.get('matrix', 'newcmds'),
        'anonymous', 'sysop',))
    cfg_bbs.set('nua', 'invalid_handles', invalid_handles)
    cfg_bbs.set('nua', 'handle_validation', '^[A-Za-z0-9]{3,11}$')

    cfg_bbs.add_section('msg')
    cfg_bbs.set('msg', 'max_subject', '40')
    # by default, anybody can make up a new tag. otherwise, only
    # those of the groups specified may.
    cfg_bbs.set('msg', 'moderated_tags', 'no')
    cfg_bbs.set('msg', 'tag_moderators', 'sysop, moderator')

    return cfg_bbs


def init_log_ini():
    """ Return ConfigParser instance of logger defaults. """
    cfg_log = ConfigParser.RawConfigParser()
    cfg_log.add_section('formatters')
    cfg_log.set('formatters', 'keys', 'default')

    cfg_log.add_section('formatter_default')
    # for multiprocessing/threads, use: %(processName)s %(threadName) !
    cfg_log.set('formatter_default', 'format',
                u'%(asctime)s %(levelname)-6s '
                u'%(filename)10s:%(lineno)-3s %(message)s')
    cfg_log.set('formatter_default', 'class', 'logging.Formatter')
    cfg_log.set('formatter_default', 'datefmt', '%a-%m-%d %I:%M%p')

    cfg_log.add_section('handlers')
    cfg_log.set('handlers', 'keys', 'console, rotate_daily')

    cfg_log.add_section('handler_console')
    cfg_log.set('handler_console', 'class', 'logging.StreamHandler')
    cfg_log.set('handler_console', 'formatter', 'default')
    cfg_log.set('handler_console', 'args', 'tuple()')

    cfg_log.add_section('handler_rotate_daily')
    cfg_log.set('handler_rotate_daily', 'class',
                'logging.handlers.TimedRotatingFileHandler')
    cfg_log.set('handler_rotate_daily', 'level', 'INFO')
    cfg_log.set('handler_rotate_daily', 'suffix', '%Y%m%d')
    cfg_log.set('handler_rotate_daily', 'encoding', 'utf8')
    cfg_log.set('handler_rotate_daily', 'formatter', 'default')
    daily_log = os.path.join(os.path.expanduser(
        os.path.join('~', '.x84', 'daily.log')))
    cfg_log.set('handler_rotate_daily', 'args',
                '("' + daily_log + '", "midnight", 1, 60)')

    cfg_log.add_section('loggers')
    cfg_log.set('loggers', 'keys',
                'root, sqlitedict, paramiko, xmodem, requests, irc')

    cfg_log.add_section('logger_root')
    cfg_log.set('logger_root', 'level', 'INFO')
    cfg_log.set('logger_root', 'formatter', 'default')
    cfg_log.set('logger_root', 'handlers', 'console, rotate_daily')

    # squelch sqlitedict's info, its rather long
    cfg_log.add_section('logger_sqlitedict')
    cfg_log.set('logger_sqlitedict', 'level', 'WARN')
    cfg_log.set('logger_sqlitedict', 'formatter', 'default')
    cfg_log.set('logger_sqlitedict', 'handlers', 'console, rotate_daily')
    cfg_log.set('logger_sqlitedict', 'qualname', 'sqlitedict')

    # squelch paramiko.transport info, also too verbose
    cfg_log.add_section('logger_paramiko')
    cfg_log.set('logger_paramiko', 'level', 'WARN')
    cfg_log.set('logger_paramiko', 'formatter', 'default')
    cfg_log.set('logger_paramiko', 'handlers', 'console, rotate_daily')
    cfg_log.set('logger_paramiko', 'qualname', 'paramiko.transport')

    # squelch xmodem's debug, too verbose
    cfg_log.add_section('logger_xmodem')
    cfg_log.set('logger_xmodem', 'level', 'INFO')
    cfg_log.set('logger_xmodem', 'formatter', 'default')
    cfg_log.set('logger_xmodem', 'handlers', 'console, rotate_daily')
    cfg_log.set('logger_xmodem', 'qualname', 'xmodem')

    # squelch requests to warn, too verbose
    cfg_log.add_section('logger_requests')
    cfg_log.set('logger_requests', 'level', 'WARN')
    cfg_log.set('logger_requests', 'formatter', 'default')
    cfg_log.set('logger_requests', 'handlers', 'console, rotate_daily')
    cfg_log.set('logger_requests', 'qualname', 'requests')

    # squelch irc debug, privacy-invasive
    cfg_log.add_section('logger_irc')
    cfg_log.set('logger_irc', 'level', 'INFO')
    cfg_log.set('logger_irc', 'formatter', 'default')
    cfg_log.set('logger_irc', 'handlers', 'console, rotate_daily')
    cfg_log.set('logger_irc', 'qualname', 'irc.client')

    return cfg_log


def get_ini(section=None, key=None, getter='get', split=False, splitsep=','):
    """
    Get an ini configuration of ``section`` and ``key``.

    If the option does not exist, an empty list, string, or False
    is returned -- return type decided by the given arguments.

    The ``getter`` method is 'get' by default, returning a string.
    For booleans, use ``getter='get_boolean'``.

    To return a list, use ``split=True``.
    """
    assert section is not None, section
    assert key is not None, key
    if CFG is None:
        # when building documentation, 'get_ini' at module-level
        # imports is not really an error.  However, if you're importing
        # a module that calls get_ini before the config system is
        # initialized, then you're going to get an empty value! warning!!
        stack = inspect.stack()
        caller_mod, caller_func = stack[2][1], stack[2][3]
        warnings.warn('ini system not (yet) initialized, '
                      'caller = {0}:{1}'.format(caller_mod, caller_func))
    elif CFG.has_option(section, key):
        getter = getattr(CFG, getter)
        value = getter(section, key)
        if split and hasattr(value, 'split'):
            return [_value.strip() for _value in value.split(splitsep)]
        return value
    if getter == 'getboolean':
        return False
    if split:
        return []
    return u''
